from hachoir.metadata.metadata_item import QUALITY_BEST, QUALITY_FASTEST
from hachoir.metadata.safe import fault_tolerant, getValue
from hachoir.metadata.metadata import (
    RootMetadata, Metadata, MultipleMetadata, registerExtractor)
from hachoir.parser.archive import (Bzip2Parser, CabFile, GzipParser,
                                    TarFile, ZipFile, MarFile)
from hachoir.core.tools import humanUnixAttributes


def maxNbFile(meta):
    if meta.quality <= QUALITY_FASTEST:
        return 0
    if QUALITY_BEST <= meta.quality:
        return None
    return 1 + int(10 * meta.quality)


def computeCompressionRate(meta):
    """
    Compute compression rate, sizes have to be in byte.
    """
    if (not meta.has("file_size")
            or not meta.get("compr_size", 0)):
        return
    file_size = meta.get("file_size")
    if not file_size:
        return
    meta.compr_rate = float(file_size) / meta.get("compr_size")


class Bzip2Metadata(RootMetadata):

    def extract(self, zip):
        if "file" in zip:
            self.compr_size = zip["file"].size // 8


class GzipMetadata(RootMetadata):

    def extract(self, gzip):
        self.useHeader(gzip)
        computeCompressionRate(self)

    @fault_tolerant
    def useHeader(self, gzip):
        self.compression = gzip["compression"].display
        if gzip["mtime"]:
            self.last_modification = gzip["mtime"].value
        self.os = gzip["os"].display
        if gzip["has_filename"].value:
            self.filename = getValue(gzip, "filename")
        if gzip["has_comment"].value:
            self.comment = getValue(gzip, "comment")
        self.compr_size = gzip["file"].size // 8
        self.file_size = gzip["size"].value


class ZipMetadata(MultipleMetadata):

    def extract(self, zip):
        max_nb = maxNbFile(self)
        for index, field in enumerate(zip.array("file")):
            if max_nb is not None and max_nb <= index:
                self.warning("ZIP archive contains many files, "
                             "but only first %s files are processed"
                             % max_nb)
                break
            self.processFile(field)

        self.extract_end_central_directory(zip)

    @fault_tolerant
    def extract_end_central_directory(self, parser):
        comment = parser['end_central_directory/comment'].value
        if comment:
            self.comment = comment

    @fault_tolerant
    def processFile(self, field):
        meta = Metadata(self)
        meta.filename = field["filename"].value
        meta.creation_date = field["last_mod"].value
        meta.compression = field["compression"].display
        if "data_desc" in field:
            meta.file_size = field["data_desc/file_uncompressed_size"].value
            if field["data_desc/file_compressed_size"].value:
                meta.compr_size = field["data_desc/file_compressed_size"].value
        else:
            meta.file_size = field["uncompressed_size"].value
            if field["compressed_size"].value:
                meta.compr_size = field["compressed_size"].value
        computeCompressionRate(meta)
        self.addGroup(field.name, meta, "File \"%s\"" % meta.get('filename'))


class TarMetadata(MultipleMetadata):

    def extract(self, tar):
        max_nb = maxNbFile(self)
        for index, field in enumerate(tar.array("file")):
            if max_nb is not None and max_nb <= index:
                self.warning("TAR archive contains many files, "
                             "but only first %s files are processed"
                             % max_nb)
                break
            meta = Metadata(self)
            self.extractFile(field, meta)
            if meta.has("filename"):
                title = 'File "%s"' % meta.getText('filename')
            else:
                title = "File"
            self.addGroup(field.name, meta, title)

    @fault_tolerant
    def extractFile(self, field, meta):
        meta.filename = field["name"].value
        meta.file_attr = humanUnixAttributes(field.getOctal("mode"))
        meta.file_size = field.getOctal("size")
        try:
            if field.getOctal("mtime"):
                meta.last_modification = field.getDatetime()
        except ValueError:
            pass
        meta.file_type = field["type"].display
        meta.author = "%s (uid=%s), group %s (gid=%s)" %\
            (field["uname"].value, field.getOctal("uid"),
             field["gname"].value, field.getOctal("gid"))


class CabMetadata(MultipleMetadata):

    def extract(self, cab):
        if "folder[0]" in cab:
            self.useFolder(cab["folder[0]"])
        self.format_version = ("Microsoft Cabinet version %s.%s"
                               % (cab["major_version"].display,
                                  cab["minor_version"].display))
        self.comment = "%s folders, %s files" % (
            cab["nb_folder"].value, cab["nb_files"].value)
        max_nb = maxNbFile(self)
        for index, field in enumerate(cab.array("file")):
            if max_nb is not None and max_nb <= index:
                self.warning("CAB archive contains many files, "
                             "but only first %s files are processed"
                             % max_nb)
                break
            self.useFile(field)

    @fault_tolerant
    def useFolder(self, folder):
        compr = folder["compr_method"].display
        if folder["compr_method"].value != 0:
            compr += " (level %u)" % folder["compr_level"].value
        self.compression = compr

    @fault_tolerant
    def useFile(self, field):
        meta = Metadata(self)
        meta.filename = field["filename"].value
        meta.file_size = field["filesize"].value
        meta.creation_date = field["timestamp"].value
        attr = field["attributes"].value
        if attr != "(none)":
            meta.file_attr = attr
        if meta.has("filename"):
            title = "File \"%s\"" % meta.getText('filename')
        else:
            title = "File"
        self.addGroup(field.name, meta, title)


class MarMetadata(MultipleMetadata):

    def extract(self, mar):
        self.comment = "Contains %s files" % mar["nb_file"].value
        self.format_version = ("Microsoft Archive version %s"
                               % mar["version"].value)
        max_nb = maxNbFile(self)
        for index, field in enumerate(mar.array("file")):
            if max_nb is not None and max_nb <= index:
                self.warning("MAR archive contains many files, "
                             "but only first %s files are processed"
                             % max_nb)
                break
            meta = Metadata(self)
            meta.filename = field["filename"].value
            meta.compression = "None"
            meta.file_size = field["filesize"].value
            self.addGroup(field.name, meta,
                          "File \"%s\"" % meta.getText('filename'))


registerExtractor(CabFile, CabMetadata)
registerExtractor(GzipParser, GzipMetadata)
registerExtractor(Bzip2Parser, Bzip2Metadata)
registerExtractor(TarFile, TarMetadata)
registerExtractor(ZipFile, ZipMetadata)
registerExtractor(MarFile, MarMetadata)



"""
Shared code for tracker parser.
"""

NOTE_NAME = {}
NOTES = ("C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "G#", "A", "A#", "B")
for octave in range(10):
    for index, note in enumerate(NOTES):
        NOTE_NAME[octave * 12 + index] = "%s (octave %s)" % (note, octave)



from hachoir.parser.misc.file_3do import File3do  # noqa
from hachoir.parser.misc.file_3ds import File3ds  # noqa
from hachoir.parser.misc.torrent import TorrentFile  # noqa
from hachoir.parser.misc.ttf import TrueTypeFontFile  # noqa
from hachoir.parser.misc.chm import ChmFile  # noqa
from hachoir.parser.misc.lnk import LnkFile  # noqa
from hachoir.parser.misc.pcf import PcfFile  # noqa
from hachoir.parser.misc.ole2 import OLE2_File  # noqa
from hachoir.parser.misc.pdf import PDFDocument  # noqa
from hachoir.parser.misc.pifv import PIFVFile  # noqa
from hachoir.parser.misc.hlp import HlpFile  # noqa
from hachoir.parser.misc.gnome_keyring import GnomeKeyring  # noqa
from hachoir.parser.misc.bplist import BPList  # noqa
from hachoir.parser.misc.dsstore import DSStore  # noqa
from hachoir.parser.misc.word_doc import WordDocumentParser  # noqa
from hachoir.parser.misc.word_2 import Word2DocumentParser  # noqa
from hachoir.parser.misc.mstask import MSTaskFile  # noqa
from hachoir.parser.misc.mapsforge_map import MapsforgeMapFile  # noqa
from hachoir.parser.misc.fit import FITFile  # noqa



"""
MS-DOS structures.

Documentation:
- File attributes:
  http://www.cs.colorado.edu/~main/cs1300/include/ddk/winddk.h
"""

from hachoir.field import StaticFieldSet
from hachoir.field import Bit, NullBits

_FIELDS = (
    (Bit, "read_only"),
    (Bit, "hidden"),
    (Bit, "system"),
    (NullBits, "reserved[]", 1),
    (Bit, "directory"),
    (Bit, "archive"),
    (Bit, "device"),
    (Bit, "normal"),
    (Bit, "temporary"),
    (Bit, "sparse_file"),
    (Bit, "reparse_file"),
    (Bit, "compressed"),
    (Bit, "offline"),
    (Bit, "dont_index_content"),
    (Bit, "encrypted"),
)


class MSDOSFileAttr16(StaticFieldSet):
    """
    MSDOS 16-bit file attributes
    """
    format = _FIELDS + ((NullBits, "reserved[]", 1),)

    _text_keys = (
        # Sort attributes by importance
        "directory", "read_only", "compressed",
        "hidden", "system",
        "normal", "device",
        "temporary", "archive")

    def createValue(self):
        mode = []
        for name in self._text_keys:
            if self[name].value:
                if 4 <= len(mode):
                    mode.append("...")
                    break
                else:
                    mode.append(name)
        if mode:
            return ", ".join(mode)
        else:
            return "(none)"


class MSDOSFileAttr32(MSDOSFileAttr16):
    """
    MSDOS 32-bit file attributes
    """
    format = _FIELDS + ((NullBits, "reserved[]", 17),)



"""
Photoshop parser (.psd file).

Creation date: 8 january 2006
Author: Victor Stinner
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt16, UInt32, String, NullBytes, Enum, RawBytes)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.parser.image.photoshop_metadata import Photoshop8BIM


class Config(FieldSet):

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = (4 + self["size"].value) * 8

    def createFields(self):
        yield UInt32(self, "size")
        while not self.eof:
            yield Photoshop8BIM(self, "item[]")


class PsdFile(Parser):
    endian = BIG_ENDIAN
    PARSER_TAGS = {
        "id": "psd",
        "category": "image",
        "file_ext": ("psd",),
        "mime": ("image/psd", "image/photoshop", "image/x-photoshop"),
        "min_size": 4 * 8,
        "magic": ((b"8BPS\0\1", 0),),
        "description": "Photoshop (PSD) picture",
    }
    COLOR_MODE = {
        0: "Bitmap",
        1: "Grayscale",
        2: "Indexed",
        3: "RGB color",
        4: "CMYK color",
        7: "Multichannel",
        8: "Duotone",
        9: "Lab Color",
    }
    COMPRESSION_NAME = {
        0: "Raw data",
        1: "RLE",
    }

    def validate(self):
        if self.stream.readBytes(0, 4) != b"8BPS":
            return "Invalid signature"
        return True

    def createFields(self):
        yield String(self, "signature", 4, "PSD signature (8BPS)", charset="ASCII")
        yield UInt16(self, "version")
        yield NullBytes(self, "reserved[]", 6)
        yield UInt16(self, "nb_channels")
        yield UInt32(self, "width")
        yield UInt32(self, "height")
        yield UInt16(self, "depth")
        yield Enum(UInt16(self, "color_mode"), self.COLOR_MODE)

        # Mode data
        yield UInt32(self, "mode_data_size")
        size = self["mode_data_size"].value
        if size:
            yield RawBytes(self, "mode_data", size)

        # Resources
        yield Config(self, "config")

        # Reserved
        yield UInt32(self, "reserved_data_size")
        size = self["reserved_data_size"].value
        if size:
            yield RawBytes(self, "reserved_data", size)

        yield Enum(UInt16(self, "compression"), self.COMPRESSION_NAME)

        size = (self.size - self.current_size) // 8
        if size:
            yield RawBytes(self, "end", size)



"""
ELF (Unix/BSD executable file format) parser.

Author: Victor Stinner, Robert Xiao
Creation date: 08 may 2006
Reference:
- System V Application Binary Interface - DRAFT - 10 June 2013
  http://www.sco.com/developers/gabi/latest/contents.html
"""

from hachoir.parser import HachoirParser
from hachoir.field import (RootSeekableFieldSet, FieldSet, Bit, NullBits, RawBits,
                           UInt8, UInt16, UInt32, UInt64, Enum,
                           String, RawBytes, Bytes)
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN


class ElfHeader(FieldSet):
    MAGIC = b"\x7FELF"
    LITTLE_ENDIAN_ID = 1
    BIG_ENDIAN_ID = 2
    MACHINE_NAME = {
        # e_machine, EM_ defines
        0: "No machine",
        1: "AT&T WE 32100",
        2: "SPARC",
        3: "Intel 80386",
        4: "Motorola 68000",
        5: "Motorola 88000",
        6: "Intel 80486",
        7: "Intel 80860",
        8: "MIPS I Architecture",
        9: "Amdahl UTS on System/370",
        10: "MIPS RS3000 Little-endian",
        11: "IBM RS/6000 XXX reserved",
        15: "Hewlett-Packard PA-RISC",
        16: "NCube XXX reserved",
        17: "Fujitsu VPP500",
        18: "Enhanced instruction set SPARC",
        19: "Intel 80960",
        20: "PowerPC 32-bit",
        21: "PowerPC 64-bit",
        22: "IBM S390",
        36: "NEC V800",
        37: "Fujitsu FR20",
        38: "TRW RH-32",
        39: "Motorola RCE",
        40: "Advanced RISC Machines (ARM)",
        41: "DIGITAL Alpha",
        42: "Hitachi Super-H",
        43: "SPARC Version 9",
        44: "Siemens Tricore",
        45: "Argonaut RISC Core",
        46: "Hitachi H8/300",
        47: "Hitachi H8/300H",
        48: "Hitachi H8S",
        49: "Hitachi H8/500",
        50: "Intel Merced (IA-64) Processor",
        51: "Stanford MIPS-X",
        52: "Motorola Coldfire",
        53: "Motorola MC68HC12",
        62: "Advanced Micro Devices x86-64",
        75: "DIGITAL VAX",
        36902: "used by NetBSD/alpha; obsolete",
    }
    CLASS_NAME = {
        # e_ident[EI_CLASS], ELFCLASS defines
        1: "32 bits",
        2: "64 bits"
    }
    TYPE_NAME = {
        # e_type, ET_ defines
        0: "No file type",
        1: "Relocatable file",
        2: "Executable file",
        3: "Shared object file",
        4: "Core file",
        0xFF00: "Processor-specific (0xFF00)",
        0xFFFF: "Processor-specific (0xFFFF)",
    }
    OSABI_NAME = {
        # e_ident[EI_OSABI], ELFOSABI_ defines
        0: "UNIX System V ABI",
        1: "HP-UX operating system",
        2: "NetBSD",
        3: "GNU/Linux",
        4: "GNU/Hurd",
        5: "86Open common IA32 ABI",
        6: "Solaris",
        7: "Monterey",
        8: "IRIX",
        9: "FreeBSD",
        10: "TRU64 UNIX",
        11: "Novell Modesto",
        12: "OpenBSD",
        97: "ARM",
        255: "Standalone (embedded) application",
    }
    ENDIAN_NAME = {
        # e_ident[EI_DATA], ELFDATA defines
        LITTLE_ENDIAN_ID: "Little endian",
        BIG_ENDIAN_ID: "Big endian",
    }

    def createFields(self):
        yield Bytes(self, "signature", 4, r'ELF signature ("\x7fELF")')
        yield Enum(UInt8(self, "class", "Class"), self.CLASS_NAME)
        if self["class"].value == 1:
            ElfLongWord = UInt32
        else:
            ElfLongWord = UInt64
        yield Enum(UInt8(self, "endian", "Endian"), self.ENDIAN_NAME)
        yield UInt8(self, "file_version", "File version")
        yield Enum(UInt8(self, "osabi_ident", "OS/syscall ABI identification"), self.OSABI_NAME)
        yield UInt8(self, "abi_version", "syscall ABI version")
        yield String(self, "pad", 7, "Pad")

        yield Enum(UInt16(self, "type", "File type"), self.TYPE_NAME)
        yield Enum(UInt16(self, "machine", "Machine type"), self.MACHINE_NAME)
        yield UInt32(self, "version", "ELF format version")
        yield textHandler(ElfLongWord(self, "entry", "Entry point"), hexadecimal)
        yield ElfLongWord(self, "phoff", "Program header file offset")
        yield ElfLongWord(self, "shoff", "Section header file offset")
        yield UInt32(self, "flags", "Architecture-specific flags")
        yield UInt16(self, "ehsize", "Elf header size (this header)")
        yield UInt16(self, "phentsize", "Program header entry size")
        yield UInt16(self, "phnum", "Program header entry count")
        yield UInt16(self, "shentsize", "Section header entry size")
        yield UInt16(self, "shnum", "Section header entry count")
        yield UInt16(self, "shstrndx", "Section header string table index")

    def isValid(self):
        if self["signature"].value != self.MAGIC:
            return "Wrong ELF signature"
        if self["class"].value not in self.CLASS_NAME:
            return "Unknown class"
        if self["endian"].value not in self.ENDIAN_NAME:
            return "Unknown endian (%s)" % self["endian"].value
        return ""


class SectionFlags(FieldSet):

    def createFields(self):
        field_thunks = (
            lambda: Bit(self, "is_writable", "Section contains writable data?"),
            lambda: Bit(self, "is_alloc", "Section occupies memory?"),
            lambda: Bit(self, "is_exec", "Section contains executable instructions?"),
            lambda: NullBits(self, "reserved[]", 1),
            lambda: Bit(self, "is_merged", "Section might be merged to eliminate duplication?"),
            lambda: Bit(self, "is_strings", "Section contains nul terminated strings?"),
            lambda: Bit(self, "is_info_link", "sh_info field of this section header holds section header table index?"),
            lambda: Bit(self, "preserve_link_order", "Section requires special ordering for linker?"),
            lambda: Bit(self, "os_nonconforming", "Section rqeuires OS-specific processing?"),
            lambda: Bit(self, "is_group", "Section is a member of a section group?"),
            lambda: Bit(self, "is_tls", "Section contains TLS data?"),
            lambda: Bit(self, "is_compressed", "Section contains compressed data?"),
            lambda: NullBits(self, "reserved[]", 8),
            lambda: RawBits(self, "os_specific", 8, "OS specific flags"),
            lambda: RawBits(self, "processor_specific", 4, "Processor specific flags"),
        )

        if self.root.endian == BIG_ENDIAN:
            if self.root.is64bit:
                yield RawBits(self, "reserved[]", 32)
            for t in reversed(field_thunks):
                yield t()
        else:
            for t in field_thunks:
                yield t()
            if self.root.is64bit:
                yield RawBits(self, "reserved[]", 32)


class SymbolStringTableOffset(UInt32):

    def createDisplay(self):
        section_index = self['/header/shstrndx'].value
        section = self['/section[' + str(section_index) + ']']
        text = section.value[self.value:]
        text = text.decode('utf-8')
        return text.split('\0', 1)[0]


class SectionHeader32(FieldSet):
    static_size = 40 * 8
    TYPE_NAME = {
        # sh_type, SHT_ defines
        0: "Inactive",
        1: "Program defined information",
        2: "Symbol table section",
        3: "String table section",
        4: "Relocation section with addends",
        5: "Symbol hash table section",
        6: "Dynamic section",
        7: "Note section",
        8: "Block started by symbol (BSS) or No space section",
        9: "Relocation section without addends",
        10: "Reserved - purpose unknown",
        11: "Dynamic symbol table section",
    }

    def createFields(self):
        yield SymbolStringTableOffset(self, "name", "Section name (index into section header string table)")
        yield Enum(textHandler(UInt32(self, "type", "Section type"), hexadecimal), self.TYPE_NAME)
        yield SectionFlags(self, "flags", "Section flags")
        yield textHandler(UInt32(self, "VMA", "Virtual memory address"), hexadecimal)
        yield textHandler(UInt32(self, "LMA", "Logical memory address (offset in file)"), hexadecimal)
        yield textHandler(UInt32(self, "size", "Section size (bytes)"), hexadecimal)
        yield UInt32(self, "link", "Index of a related section")
        yield UInt32(self, "info", "Type-dependent information")
        yield UInt32(self, "addr_align", "Address alignment (bytes)")
        yield UInt32(self, "entry_size", "Size of each entry in section")

    def createDescription(self):
        return "Section header (name: %s, type: %s)" % \
            (self["name"].display, self["type"].display)


class SectionHeader64(SectionHeader32):
    static_size = 64 * 8

    def createFields(self):
        yield SymbolStringTableOffset(self, "name", "Section name (index into section header string table)")
        yield Enum(textHandler(UInt32(self, "type", "Section type"), hexadecimal), self.TYPE_NAME)
        yield SectionFlags(self, "flags", "Section flags")
        yield textHandler(UInt64(self, "VMA", "Virtual memory address"), hexadecimal)
        yield textHandler(UInt64(self, "LMA", "Logical memory address (offset in file)"), hexadecimal)
        yield textHandler(UInt64(self, "size", "Section size (bytes)"), hexadecimal)
        yield UInt32(self, "link", "Index of a related section")
        yield UInt32(self, "info", "Type-dependent information")
        yield UInt64(self, "addr_align", "Address alignment (bytes)")
        yield UInt64(self, "entry_size", "Size of each entry in section")


class ProgramFlags(FieldSet):
    static_size = 32
    FLAGS = (('pf_r', 'readable'), ('pf_w', 'writable'), ('pf_x', 'executable'))

    def createFields(self):
        if self.root.endian == BIG_ENDIAN:
            yield NullBits(self, "padding[]", 29)
            for fld, desc in self.FLAGS:
                yield Bit(self, fld, "Segment is " + desc)
        else:
            for fld, desc in reversed(self.FLAGS):
                yield Bit(self, fld, "Segment is " + desc)
            yield NullBits(self, "padding[]", 29)

    def createDescription(self):
        attribs = []
        for fld, desc in self.FLAGS:
            if self[fld].value:
                attribs.append(desc)
        return 'Segment is ' + ', '.join(attribs)


class ProgramHeader32(FieldSet):
    TYPE_NAME = {
        # p_type, PT_ defines
        0: "Unused program header table entry",
        1: "Loadable program segment",
        2: "Dynamic linking information",
        3: "Program interpreter",
        4: "Auxiliary information",
        5: "Reserved, unspecified semantics",
        6: "Entry for header table itself",
        7: "Thread Local Storage segment",
        0x70000000: "MIPS_REGINFO",
    }
    static_size = 32 * 8

    def createFields(self):
        yield Enum(UInt32(self, "type", "Segment type"), ProgramHeader32.TYPE_NAME)
        yield UInt32(self, "offset", "Offset")
        yield textHandler(UInt32(self, "vaddr", "V. address"), hexadecimal)
        yield textHandler(UInt32(self, "paddr", "P. address"), hexadecimal)
        yield UInt32(self, "file_size", "File size")
        yield UInt32(self, "mem_size", "Memory size")
        yield ProgramFlags(self, "flags")
        yield UInt32(self, "align", "Alignment padding")

    def createDescription(self):
        return "Program Header (%s)" % self["type"].display


class ProgramHeader64(ProgramHeader32):
    static_size = 56 * 8

    def createFields(self):
        yield Enum(UInt32(self, "type", "Segment type"), ProgramHeader32.TYPE_NAME)
        yield ProgramFlags(self, "flags")
        yield UInt64(self, "offset", "Offset")
        yield textHandler(UInt64(self, "vaddr", "V. address"), hexadecimal)
        yield textHandler(UInt64(self, "paddr", "P. address"), hexadecimal)
        yield UInt64(self, "file_size", "File size")
        yield UInt64(self, "mem_size", "Memory size")
        yield UInt64(self, "align", "Alignment padding")


class ElfFile(HachoirParser, RootSeekableFieldSet):
    PARSER_TAGS = {
        "id": "elf",
        "category": "program",
        "file_ext": ("so", ""),
        "min_size": 52 * 8,  # At least one program header
        "mime": (
            "application/x-executable",
            "application/x-object",
            "application/x-sharedlib",
            "application/x-executable-file",
            "application/x-coredump"),
        "magic": ((ElfHeader.MAGIC, 0),),
        "description": "ELF Unix/BSD program/library"
    }
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, len(ElfHeader.MAGIC)) != ElfHeader.MAGIC:
            return "Invalid magic"
        err = self["header"].isValid()
        if err:
            return err
        return True

    def createFields(self):
        # Choose the right endian depending on endian specified in header
        if self.stream.readBits(5 * 8, 8, BIG_ENDIAN) == ElfHeader.BIG_ENDIAN_ID:
            self.endian = BIG_ENDIAN
        else:
            self.endian = LITTLE_ENDIAN

        # Parse header and program headers
        yield ElfHeader(self, "header", "Header")
        self.is64bit = (self["header/class"].value == 2)

        for index in range(self["header/phnum"].value):
            if self.is64bit:
                yield ProgramHeader64(self, "prg_header[]")
            else:
                yield ProgramHeader32(self, "prg_header[]")

        self.seekByte(self["header/shoff"].value, relative=False)

        for index in range(self["header/shnum"].value):
            if self.is64bit:
                yield SectionHeader64(self, "section_header[]")
            else:
                yield SectionHeader32(self, "section_header[]")

        for index in range(self["header/shnum"].value):
            field = self["section_header[" + str(index) + "]"]
            if field['size'].value != 0 and field['type'].value != 8:
                # skip NOBITS sections
                self.seekByte(field['LMA'].value, relative=False)
                yield RawBytes(self, "section[" + str(index) + "]", field['size'].value)

    def createDescription(self):
        return "ELF Unix/BSD program/library: %s" % (
            self["header/class"].display)



import wx


class frame_view_fwd_t:
    def __init__(self, imp):
        self.imp = imp

    def on_frame_view_ready(self, dispatcher, view):
        assert view is not None
        view.Bind(wx.EVT_ACTIVATE, self.on_activated)
        view.Bind(wx.EVT_SHOW, self.on_shown)

    def on_activated(self, event):
        if event.GetActive():
            self.imp.on_activated()

    def on_shown(self, event):
        if event.Show:
            self.imp.on_activated()



from hachoir.field import (
    RawBits, Bit, Bits, PaddingBits,
    RawBytes, Bytes, PaddingBytes,
    GenericString, Character,
    TimestampMac32,
    isInteger, isString)
from hachoir.core.tools import timestampMac32
from hachoir.editor import FakeField
from datetime import datetime


class EditableField(FakeField):
    """
    Pure virtual class used to write editable field class.
    """

    _is_altered = False

    def __init__(self, parent, name, value=None):
        FakeField.__init__(self, parent, name)
        self._value = value

    def _isAltered(self):
        return self._is_altered
    is_altered = property(_isAltered)

    def hasValue(self):
        return True

    def _computeSize(self):
        raise NotImplementedError()

    def _getValue(self):
        return self._value

    def _setValue(self, value):
        self._value = value

    def _propGetValue(self):
        if self._value is not None:
            return self._getValue()
        else:
            return FakeField._getValue(self)

    def _propSetValue(self, value):
        self._setValue(value)
        self._is_altered = True
    value = property(_propGetValue, _propSetValue)

    def _getSize(self):
        if self._value is not None:
            return self._computeSize()
        else:
            return FakeField._getSize(self)
    size = property(_getSize)

    def _write(self, output):
        raise NotImplementedError()

    def writeInto(self, output):
        if self._is_altered:
            self._write(output)
        else:
            return FakeField.writeInto(self, output)


class EditableFixedField(EditableField):
    """
    Editable field with fixed size.
    """

    def __init__(self, parent, name, value=None, size=None):
        EditableField.__init__(self, parent, name, value)
        if size is not None:
            self._size = size
        else:
            self._size = self._parent._getOriginalField(self._name).size

    def _getSize(self):
        return self._size
    size = property(_getSize)


class EditableBits(EditableFixedField):

    def __init__(self, parent, name, *args):
        if args:
            if len(args) != 2:
                raise TypeError(
                    "Wrong argument count, EditableBits constructor prototype is: "
                    "(parent, name, [size, value])")
            size = args[0]
            value = args[1]
            assert isinstance(value, int)
        else:
            size = None
            value = None
        EditableFixedField.__init__(self, parent, name, value, size)
        if args:
            self._setValue(args[1])
            self._is_altered = True

    def _setValue(self, value):
        if not (0 <= value < (1 << self._size)):
            raise ValueError("Invalid value, must be in range %s..%s"
                             % (0, (1 << self._size) - 1))
        self._value = value

    def _write(self, output):
        output.writeBits(self._size, self._value, self._parent.endian)


class EditableBytes(EditableField):

    def _setValue(self, value):
        if not value:
            raise ValueError(
                "Unable to set empty string to a EditableBytes field")
        self._value = value

    def _computeSize(self):
        return len(self._value) * 8

    def _write(self, output):
        output.writeBytes(self._value)


class EditableString(EditableField):
    MAX_SIZE = {
        "Pascal8": (1 << 8) - 1,
        "Pascal16": (1 << 16) - 1,
        "Pascal32": (1 << 32) - 1,
    }

    def __init__(self, parent, name, *args, **kw):
        if len(args) == 2:
            value = args[1]
            assert isinstance(value, bytes)  # TODO: support Unicode
        elif not args:
            value = None
        else:
            raise TypeError(
                "Wrong argument count, EditableString constructor prototype is:"
                "(parent, name, [format, value])")
        EditableField.__init__(self, parent, name, value)
        if len(args) == 2:
            self._charset = kw.get('charset', None)
            self._format = args[0]
            if self._format in GenericString.PASCAL_FORMATS:
                self._prefix_size = GenericString.PASCAL_FORMATS[self._format]
            else:
                self._prefix_size = 0
            self._suffix_str = GenericString.staticSuffixStr(
                self._format, self._charset, self._parent.endian)
            self._is_altered = True
        else:
            orig = self._parent._getOriginalField(name)
            self._charset = orig.charset
            self._format = orig.format
            self._prefix_size = orig.content_offset
            self._suffix_str = orig.suffix_str

    def _setValue(self, value):
        size = len(value)
        if self._format in self.MAX_SIZE and self.MAX_SIZE[self._format] < size:
            raise ValueError("String is too big")
        assert isinstance(value, bytes)
        self._value = value

    def _computeSize(self):
        return (self._prefix_size + len(self._value) + len(self._suffix_str)) * 8

    def _write(self, output):
        if self._format in GenericString.SUFFIX_FORMAT:
            output.writeBytes(self._value)
            output.writeBytes(self._suffix_str)
        elif self._format == "fixed":
            output.writeBytes(self._value)
        else:
            assert self._format in GenericString.PASCAL_FORMATS
            size = GenericString.PASCAL_FORMATS[self._format]
            output.writeInteger(len(self._value), False,
                                size, self._parent.endian)
            value = self._value.encode(self._charset)
            output.writeBytes(value)


class EditableCharacter(EditableFixedField):

    def __init__(self, parent, name, *args):
        if args:
            if len(args) != 3:
                raise TypeError(
                    "Wrong argument count, EditableCharacter "
                    "constructor prototype is: (parent, name, [value])")
            value = args[0]
            if not isinstance(value, str) or len(value) != 1:
                raise TypeError("EditableCharacter needs a character")
        else:
            value = None
        EditableFixedField.__init__(self, parent, name, value, 8)
        if args:
            self._is_altered = True

    def _setValue(self, value):
        if not isinstance(value, str) or len(value) != 1:
            raise TypeError("EditableCharacter needs a character")
        self._value = value

    def _write(self, output):
        output.writeBytes(self._value)


class EditableInteger(EditableFixedField):
    VALID_VALUE_SIGNED = {
        8: (-(1 << 8), (1 << 8) - 1),
        16: (-(1 << 15), (1 << 15) - 1),
        32: (-(1 << 31), (1 << 31) - 1),
    }
    VALID_VALUE_UNSIGNED = {
        8: (0, (1 << 8) - 1),
        16: (0, (1 << 16) - 1),
        32: (0, (1 << 32) - 1)
    }

    def __init__(self, parent, name, *args):
        if args:
            if len(args) != 3:
                raise TypeError(
                    "Wrong argument count, EditableInteger constructor prototype is: "
                    "(parent, name, [signed, size, value])")
            size = args[1]
            value = args[2]
            assert isinstance(value, int)
        else:
            size = None
            value = None
        EditableFixedField.__init__(self, parent, name, value, size)
        if args:
            self._signed = args[0]
            self._is_altered = True
        else:
            self._signed = self._parent._getOriginalField(self._name).signed

    def _setValue(self, value):
        if self._signed:
            valid = self.VALID_VALUE_SIGNED
        else:
            valid = self.VALID_VALUE_UNSIGNED
        minval, maxval = valid[self._size]
        if not (minval <= value <= maxval):
            raise ValueError("Invalid value, must be in range %s..%s"
                             % (minval, maxval))
        self._value = value

    def _write(self, output):
        output.writeInteger(
            self.value, self._signed, self._size // 8, self._parent.endian)


class EditableTimestampMac32(EditableFixedField):
    minval, maxval = timestampMac32(0), timestampMac32(2**32 - 1)

    def __init__(self, parent, name, *args):
        if args:
            if len(args) != 3:
                raise TypeError(
                    "Wrong argument count, EditableTimestampMac32 constructor prototype is: "
                    "(parent, name, [value])")
            value = args[0]
            assert isinstance(value, datetime)
        else:
            value = None
        EditableFixedField.__init__(self, parent, name, value, 32)

    def _setValue(self, value):
        if not (self.minval <= value <= self.maxval):
            raise ValueError("Invalid value, must be in range %s..%s"
                             % (self.minval, self.maxval))
        self._value = value

    def _write(self, output):
        timestamp = int((self._value - self.minval).total_seconds())
        output.writeBits(self._size, timestamp, self._parent.endian)


def createEditableField(fieldset, field):
    if isInteger(field):
        cls = EditableInteger
    elif isString(field):
        cls = EditableString
    elif field.__class__ == TimestampMac32:
        cls = EditableTimestampMac32
    elif field.__class__ in (RawBytes, Bytes, PaddingBytes):
        cls = EditableBytes
    elif field.__class__ in (RawBits, Bits, Bit, PaddingBits):
        cls = EditableBits
    elif field.__class__ == Character:
        cls = EditableCharacter
    else:
        cls = FakeField
    return cls(fieldset, field.name)



"""
AMF metadata (inside Flash video, FLV file) parser.

Documentation:

 - flashticle: Python project to read Flash (formats SWF, FLV and AMF)
   http://undefined.org/python/#flashticle

Author: Victor Stinner
Creation date: 4 november 2006
"""

from hachoir.field import (FieldSet, ParserError,
                           UInt8, UInt16, UInt32, PascalString16, Float64)
from hachoir.core.tools import timestampUNIX


def parseUTF8(parent):
    yield PascalString16(parent, "value", charset="UTF-8")


def parseDouble(parent):
    yield Float64(parent, "value")


def parseBool(parent):
    yield UInt8(parent, "value")


def parseArray(parent):
    yield UInt32(parent, "count")
    for index in range(parent["count"].value):
        yield AMFObject(parent, "item[]")


def parseObjectAttributes(parent):
    while True:
        item = Attribute(parent, "attr[]")
        yield item
        if item["key"].value == "":
            break


def parseMixedArray(parent):
    yield UInt32(parent, "count")
    for index in range(parent["count"].value + 1):
        item = Attribute(parent, "item[]")
        yield item
        if not item['key'].value:
            break


def parseDate(parent):
    yield Float64(parent, "timestamp_microsec")
    yield UInt16(parent, "timestamp_sec")


def parseNothing(parent):
    # Python hack to make sure that the function is a generator
    if 0:
        yield


class AMFObject(FieldSet):
    CODE_DATE = 11
    tag_info = {
        # http://osflash.org/amf/astypes
        0: (parseDouble, "Double"),
        1: (parseBool, "Boolean"),
        2: (parseUTF8, "UTF-8 string"),
        3: (parseObjectAttributes, "Object attributes"),
        # MOVIECLIP = '\x04',
        # NULL = '\x05',
        # UNDEFINED = '\x06',
        # REFERENCE = '\x07',
        8: (parseMixedArray, "Mixed array"),
        9: (parseNothing, "End of object"),
        10: (parseArray, "Array"),
        CODE_DATE: (parseDate, "Date"),
        # LONGUTF8 = '\x0c',
        # UNSUPPORTED = '\x0d',
        #  Server-to-client only
        # RECORDSET = '\x0e',
        # XML = '\x0f',
        # TYPEDOBJECT = '\x10',
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        code = self["type"].value
        try:
            self.parser, desc = self.tag_info[code]
            if code == self.CODE_DATE:
                self.createValue = self.createValueDate
        except KeyError:
            raise ParserError("AMF: Unable to parse type %s" % code)

    def createFields(self):
        yield UInt8(self, "type")
        yield from self.parser(self)

    def createValueDate(self):
        value = (self["timestamp_microsec"].value * 0.001) \
            - (self["timestamp_sec"].value * 60)
        return timestampUNIX(value)


class Attribute(AMFObject):

    def __init__(self, *args):
        AMFObject.__init__(self, *args)
        self._description = None

    def createFields(self):
        yield PascalString16(self, "key", charset="UTF-8")
        yield UInt8(self, "type")
        yield from self.parser(self)

    def createDescription(self):
        return 'Attribute "%s"' % self["key"].value



"""
Tar archive parser.

Author: Victor Stinner
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           Enum, UInt8, SubFile, String, NullBytes)
from hachoir.core.tools import humanFilesize, paddingSize, timestampUNIX
from hachoir.core.endian import BIG_ENDIAN
import re


class FileEntry(FieldSet):
    type_name = {
        # 48 is "0", 49 is "1", ...
        0: "Normal disk file (old format)",
        48: "Normal disk file",
        49: "Link to previously dumped file",
        50: "Symbolic link",
        51: "Character special file",
        52: "Block special file",
        53: "Directory",
        54: "FIFO special file",
        55: "Contiguous file"
    }

    def getOctal(self, name):
        return self.octal2int(self[name].value)

    def getDatetime(self):
        """
        Create modification date as Unicode string, may raise ValueError.
        """
        timestamp = self.getOctal("mtime")
        return timestampUNIX(timestamp)

    def createFields(self):
        yield String(self, "name", 100, "Name", strip="\0", charset="ISO-8859-1")
        yield String(self, "mode", 8, "Mode", strip=" \0", charset="ASCII")
        yield String(self, "uid", 8, "User ID", strip=" \0", charset="ASCII")
        yield String(self, "gid", 8, "Group ID", strip=" \0", charset="ASCII")
        yield String(self, "size", 12, "Size", strip=" \0", charset="ASCII")
        yield String(self, "mtime", 12, "Modification time", strip=" \0", charset="ASCII")
        yield String(self, "check_sum", 8, "Check sum", strip=" \0", charset="ASCII")
        yield Enum(UInt8(self, "type", "Type"), self.type_name)
        yield String(self, "lname", 100, "Link name", strip=" \0", charset="ISO-8859-1")
        yield String(self, "magic", 8, "Magic", strip=" \0", charset="ASCII")
        yield String(self, "uname", 32, "User name", strip=" \0", charset="ISO-8859-1")
        yield String(self, "gname", 32, "Group name", strip=" \0", charset="ISO-8859-1")
        yield String(self, "devmajor", 8, "Dev major", strip=" \0", charset="ASCII")
        yield String(self, "devminor", 8, "Dev minor", strip=" \0", charset="ASCII")
        yield String(self, "prefix", 155, "Prefix for filename", strip="\0", charset="ASCII")
        yield NullBytes(self, "padding", 12, "Padding (zero)")

        filesize = self.getOctal("size")
        if filesize:
            yield SubFile(self, "content", filesize, filename=self["name"].value)

        size = paddingSize(self.current_size // 8, 512)
        if size:
            yield NullBytes(self, "padding_end", size, "Padding (512 align)")

    def convertOctal(self, chunk):
        return self.octal2int(chunk.value)

    def isEmpty(self):
        return self["name"].value == ""

    def octal2int(self, text):
        try:
            return int(text, 8)
        except ValueError:
            return 0

    def createDescription(self):
        if self.isEmpty():
            desc = "(terminator, empty header)"
        else:
            filename = self["name"].value
            if self["prefix"].value:
                filename = self["prefix"].value + '/' + filename
            filesize = humanFilesize(self.getOctal("size"))
            desc = "(%s: %s, %s)" % \
                (filename, self["type"].display, filesize)
        return "Tar File " + desc


class TarFile(Parser):
    endian = BIG_ENDIAN
    PARSER_TAGS = {
        "id": "tar",
        "category": "archive",
        "file_ext": ("tar",),
        "mime": ("application/x-tar", "application/x-gtar"),
        "min_size": 512 * 8,
        "magic": ((b"ustar  \0", 257 * 8),),
        "subfile": "skip",
        "description": "TAR archive",
    }
    _sign = re.compile(b"ustar *\0|[ \0]*$")

    def validate(self):
        if not self._sign.match(self.stream.readBytes(257 * 8, 8)):
            return "Invalid magic number"
        if self[0].name == "terminator":
            return "Don't contain any file"
        try:
            int(self["file[0]/uid"].value, 8)
            int(self["file[0]/gid"].value, 8)
            int(self["file[0]/size"].value, 8)
        except ValueError:
            return "Invalid file size"
        return True

    def createFields(self):
        while not self.eof:
            field = FileEntry(self, "file[]")
            if field.isEmpty():
                yield NullBytes(self, "terminator", 512)
                break
            yield field
        if self.current_size < self._size:
            yield self.seekBit(self._size, "end")

    def createContentSize(self):
        return self["terminator"].address + self["terminator"].size



"""
Utilities used to convert a field to human classic reprentation of data.
"""

from hachoir.core.tools import (
    humanDuration, humanFilesize, alignValue,
    durationWin64 as doDurationWin64,
    deprecated)
from types import FunctionType, MethodType
from hachoir.field import Field


def textHandler(field, handler):
    assert isinstance(handler, (FunctionType, MethodType))
    assert issubclass(field.__class__, Field)
    field.createDisplay = lambda: handler(field)
    return field


def displayHandler(field, handler):
    assert isinstance(handler, (FunctionType, MethodType))
    assert issubclass(field.__class__, Field)
    field.createDisplay = lambda: handler(field.value)
    return field


@deprecated("Use TimedeltaWin64 field type")
def durationWin64(field):
    """
    Convert Windows 64-bit duration to string. The timestamp format is
    a 64-bit number: number of 100ns. See also timestampWin64().

    >>> durationWin64(type("", (), dict(value=2146280000, size=64)))
    '3 min 34 sec 628 ms'
    >>> durationWin64(type("", (), dict(value=(1 << 64)-1, size=64)))
    '58494 years 88 days 5 hours'
    """
    assert hasattr(field, "value") and hasattr(field, "size")
    assert field.size == 64
    delta = doDurationWin64(field.value)
    return humanDuration(delta)


def filesizeHandler(field):
    """
    Format field value using humanFilesize()
    """
    return displayHandler(field, humanFilesize)


def hexadecimal(field):
    """
    Convert an integer to hexadecimal in lower case. Returns unicode string.

    >>> hexadecimal(type("", (), dict(value=412, size=16)))
    '0x019c'
    >>> hexadecimal(type("", (), dict(value=0, size=32)))
    '0x00000000'
    """
    assert hasattr(field, "value") and hasattr(field, "size")
    size = field.size
    padding = alignValue(size, 4) // 4
    pattern = "0x%%0%ux" % padding
    return pattern % field.value



"""
Parallel Realities Starfighter .pak file parser

See http://www.parallelrealities.co.uk/projects/starfighter.php
or svn://svn.debian.org/svn/pkg-games/packages/trunk/starfighter/

Author: Oliver Gerlich
"""

from hachoir.parser import Parser
from hachoir.field import (UInt32, String, SubFile, FieldSet)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import filesizeHandler


class FileEntry(FieldSet):

    def createFields(self):
        yield String(self, "filename", 56, truncate="\0")
        yield filesizeHandler(UInt32(self, "size"))
        yield SubFile(self, "data", self["size"].value, filename=self["filename"].value)

    def createDescription(self):
        return self["filename"].value


class PRSPakFile(Parser):

    PARSER_TAGS = {
        "id": "prs_pak",
        "category": "archive",
        "file_ext": ("pak",),
        "mime": (u"application/octet-stream",),
        "min_size": 4 * 8,  # just the identifier
        "magic": ((b'PACK', 0),),
        "description": "Parallel Realities Starfighter .pak archive",
    }

    endian = LITTLE_ENDIAN

    def validate(self):
        return (self.stream.readBytes(0, 4) == b'PACK'
                and self["file[0]/size"].value >= 0
                and len(self["file[0]/filename"].value) > 0)

    def createFields(self):
        yield String(self, "magic", 4)

        # all remaining data must be file entries:
        while self.current_size < self._size:
            yield FileEntry(self, "file[]")



"""
Windows 2000 - List of Locale IDs and Language Groups

Original data table:
http://www.microsoft.com/globaldev/reference/win2k/setup/lcid.mspx
"""

LANGUAGE_ID = {
    0x0436: "Afrikaans",
    0x041c: "Albanian",
    0x0401: "Arabic Saudi Arabia",
    0x0801: "Arabic Iraq",
    0x0c01: "Arabic Egypt",
    0x1001: "Arabic Libya",
    0x1401: "Arabic Algeria",
    0x1801: "Arabic Morocco",
    0x1c01: "Arabic Tunisia",
    0x2001: "Arabic Oman",
    0x2401: "Arabic Yemen",
    0x2801: "Arabic Syria",
    0x2c01: "Arabic Jordan",
    0x3001: "Arabic Lebanon",
    0x3401: "Arabic Kuwait",
    0x3801: "Arabic UAE",
    0x3c01: "Arabic Bahrain",
    0x4001: "Arabic Qatar",
    0x042b: "Armenian",
    0x042c: "Azeri Latin",
    0x082c: "Azeri Cyrillic",
    0x042d: "Basque",
    0x0423: "Belarusian",
    0x0402: "Bulgarian",
    0x0403: "Catalan",
    0x0404: "Chinese Taiwan",
    0x0804: "Chinese PRC",
    0x0c04: "Chinese Hong Kong",
    0x1004: "Chinese Singapore",
    0x1404: "Chinese Macau",
    0x041a: "Croatian",
    0x0405: "Czech",
    0x0406: "Danish",
    0x0413: "Dutch Standard",
    0x0813: "Dutch Belgian",
    0x0409: "English United States",
    0x0809: "English United Kingdom",
    0x0c09: "English Australian",
    0x1009: "English Canadian",
    0x1409: "English New Zealand",
    0x1809: "English Irish",
    0x1c09: "English South Africa",
    0x2009: "English Jamaica",
    0x2409: "English Caribbean",
    0x2809: "English Belize",
    0x2c09: "English Trinidad",
    0x3009: "English Zimbabwe",
    0x3409: "English Philippines",
    0x0425: "Estonian",
    0x0438: "Faeroese",
    0x0429: "Farsi",
    0x040b: "Finnish",
    0x040c: "French Standard",
    0x080c: "French Belgian",
    0x0c0c: "French Canadian",
    0x100c: "French Swiss",
    0x140c: "French Luxembourg",
    0x180c: "French Monaco",
    0x0437: "Georgian",
    0x0407: "German Standard",
    0x0807: "German Swiss",
    0x0c07: "German Austrian",
    0x1007: "German Luxembourg",
    0x1407: "German Liechtenstein",
    0x0408: "Greek",
    0x040d: "Hebrew",
    0x0439: "Hindi",
    0x040e: "Hungarian",
    0x040f: "Icelandic",
    0x0421: "Indonesian",
    0x0410: "Italian Standard",
    0x0810: "Italian Swiss",
    0x0411: "Japanese",
    0x043f: "Kazakh",
    0x0457: "Konkani",
    0x0412: "Korean",
    0x0426: "Latvian",
    0x0427: "Lithuanian",
    0x042f: "Macedonian",
    0x043e: "Malay Malaysia",
    0x083e: "Malay Brunei Darussalam",
    0x044e: "Marathi",
    0x0414: "Norwegian Bokmal",
    0x0814: "Norwegian Nynorsk",
    0x0415: "Polish",
    0x0416: "Portuguese Brazilian",
    0x0816: "Portuguese Standard",
    0x0418: "Romanian",
    0x0419: "Russian",
    0x044f: "Sanskrit",
    0x081a: "Serbian Latin",
    0x0c1a: "Serbian Cyrillic",
    0x041b: "Slovak",
    0x0424: "Slovenian",
    0x040a: "Spanish Traditional Sort",
    0x080a: "Spanish Mexican",
    0x0c0a: "Spanish Modern Sort",
    0x100a: "Spanish Guatemala",
    0x140a: "Spanish Costa Rica",
    0x180a: "Spanish Panama",
    0x1c0a: "Spanish Dominican Republic",
    0x200a: "Spanish Venezuela",
    0x240a: "Spanish Colombia",
    0x280a: "Spanish Peru",
    0x2c0a: "Spanish Argentina",
    0x300a: "Spanish Ecuador",
    0x340a: "Spanish Chile",
    0x380a: "Spanish Uruguay",
    0x3c0a: "Spanish Paraguay",
    0x400a: "Spanish Bolivia",
    0x440a: "Spanish El Salvador",
    0x480a: "Spanish Honduras",
    0x4c0a: "Spanish Nicaragua",
    0x500a: "Spanish Puerto Rico",
    0x0441: "Swahili",
    0x041d: "Swedish",
    0x081d: "Swedish Finland",
    0x0449: "Tamil",
    0x0444: "Tatar",
    0x041e: "Thai",
    0x041f: "Turkish",
    0x0422: "Ukrainian",
    0x0420: "Urdu",
    0x0443: "Uzbek Latin",
    0x0843: "Uzbek Cyrillic",
    0x042a: "Vietnamese",
}



"""
EXT2 (Linux) file system parser.

Author: Victor Stinner

Sources:
- EXT2FS source code
  http://ext2fsd.sourceforge.net/
- Analysis of the Ext2fs structure
  http://www.nondot.org/sabre/os/files/FileSystems/ext2fs/
- Ext4 Disk Layout
  https://ext4.wiki.kernel.org/index.php/Ext4_Disk_Layout
"""

from hachoir.parser import HachoirParser, Parser
from hachoir.field import (RootSeekableFieldSet, SeekableFieldSet, FieldSet, ParserError,
                           Bit, Bits, UInt8, UInt16, UInt32,
                           Enum, String, TimestampUnix32, RawBytes,
                           NullBytes, PaddingBits, PaddingBytes, FragmentGroup, CustomFragment)
from hachoir.core.tools import (humanDuration, humanFilesize)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler
from .linux_swap import UUID


class DirectoryEntry(FieldSet):
    file_type = {
        1: "Regular",
        2: "Directory",
        3: "Char. dev.",
        4: "Block dev.",
        5: "Fifo",
        6: "Socket",
        7: "Symlink",
        8: "Max"
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["rec_len"].value * 8

    def createFields(self):
        yield UInt32(self, "inode", "Inode")
        yield UInt16(self, "rec_len", "Record length")
        yield UInt8(self, "name_len", "Name length")
        yield Enum(UInt8(self, "file_type", "File type"), self.file_type)
        if self["name_len"].value > 0:
            yield String(self, "name", self["name_len"].value, "File name")
        size = (self._size - self.current_size) // 8
        if size:
            yield NullBytes(self, "padding", size)

    def createDescription(self):
        name = None
        if self["name_len"].value > 0:
            name = self["name"].value.strip("\0")
        if name:
            return "Directory entry: %s" % name
        else:
            return "Directory entry (empty)"


class Flags(FieldSet):

    def createValue(self):
        return self.stream.readBits(self.absolute_address, self.size, self.parent.endian)

    def createDisplay(self):
        out = [
            field.name for field in self if field.value and isinstance(field, Bit)]
        return '{' + ', '.join(out) + '}'


class FileMode(Flags):
    static_size = 16

    file_type = {
        1: "Fifo",
        2: "Character device",
        4: "Directory",
        6: "Block device",
        8: "Regular",
        10: "Symbolic link",
        12: "Socket",
    }
    file_type_letter = {
        1: "p",
        2: "c",
        4: "d",
        6: "b",
        10: "l",
        12: "s",
    }

    def createFields(self):
        yield Bit(self, "other_exec")
        yield Bit(self, "other_write")
        yield Bit(self, "other_read")
        yield Bit(self, "group_exec")
        yield Bit(self, "group_write")
        yield Bit(self, "group_read")
        yield Bit(self, "owner_exec")
        yield Bit(self, "owner_write")
        yield Bit(self, "owner_read")
        yield Bit(self, "sticky")
        yield Bit(self, "setgid")
        yield Bit(self, "setuid")
        yield Enum(Bits(self, "file_type", 4), self.file_type)

    def createDisplay(self):
        names = (
            ("owner_read", "owner_write", "owner_exec"),
            ("group_read", "group_write", "group_exec"),
            ("other_read", "other_write", "other_exec"))
        letters = "rwx"
        mode = ["-" for index in range(10)]
        index = 1
        for loop in range(3):
            for name, letter in zip(names[loop], letters):
                if self[name].value:
                    mode[index] = letter
                index += 1
        if self['sticky'].value:
            mode[9] = 'Tt'[self['other_exec'].value]
        if self['setgid'].value:
            mode[6] = 'Ss'[self['group_exec'].value]
        if self['setuid'].value:
            mode[3] = 'Ss'[self['owner_exec'].value]

        file_type = self["file_type"].value
        if file_type in self.file_type_letter:
            mode[0] = self.file_type_letter[file_type]
        return "".join(mode)


class InodeFlags(Flags):
    static_size = 32

    def createFields(self):
        yield Bit(self, "secrm", "This file requires secure deletion (not implemented)")
        yield Bit(self, "unrm", "This file should be preserved, should undeletion be desired (not implemented)")
        yield Bit(self, "compr", "File is compressed (not really implemented)")
        yield Bit(self, "sync", "All writes to the file must be synchronous")
        yield Bit(self, "immutable", "File is immutable")
        yield Bit(self, "append", "File can only be appended")
        yield Bit(self, "nodump", "The dump(1) utility should not dump this file")
        yield Bit(self, "noatime", "Do not update access time")
        yield Bit(self, "dirty", "Dirty compressed file (not used)")
        yield Bit(self, "comprblk", "File has one or more compressed clusters (not used)")
        yield Bit(self, "nocompr", "Do not compress file (not used)")
        yield Bit(self, "encrypt", "Encrypted inode")
        yield Bit(self, "index", "Directory has hashed indexes")
        yield Bit(self, "imagic", "AFS magic directory")
        yield Bit(self, "journal_data", "File data must always be written through the journal")
        yield Bit(self, "notail", "File tail should not be merged (not used by ext4)")
        yield Bit(self, "dirsync", "All directory entry data should be written synchronously")
        yield Bit(self, "topdir", "Top of directory hierarchy")
        yield Bit(self, "huge_file", "This is a huge file")
        yield Bit(self, "extents", "Inode uses extents")
        yield Bit(self, "reserved[]")
        yield Bit(self, "ea_inode", "Inode used for a large extended attribute")
        yield Bit(self, "eofblocks", "This file has blocks allocated past EOF (deprecated)")
        yield Bit(self, "reserved[]")
        yield Bit(self, "snapfile", "Inode is a snapshot (not in mainline)")
        yield Bit(self, "reserved[]")
        yield Bit(self, "snapfile_deleted", "Snapshot is being deleted (not in mainline)")
        yield Bit(self, "snapfile_shrunk", "Snapshot shrink has completed (not in mainline)")
        yield Bit(self, "inline_data", "Inode has inline data")
        yield Bit(self, "projinherit", "Create with parents projid")
        yield Bit(self, "reserved[]")
        yield Bit(self, "reserved[]", "Reserved for ext4 library")


class ExtentNode(FieldSet):

    def __init__(self, parent, name, size=None):
        FieldSet.__init__(self, parent, name, size=size)

    def createFields(self):
        yield UInt16(self, "magic")
        yield UInt16(self, "cur_entries")
        yield UInt16(self, "max_entries")
        yield UInt16(self, "tree_depth")
        yield UInt32(self, "tree_generation")

        for i in range(self['cur_entries'].value):
            yield UInt32(self, "logical_block[]")
            yield UInt16(self, "extent_length[]")
            yield UInt16(self, "physical_block_upper[]")
            yield UInt32(self, "physical_block_lower[]")


class Inode(FieldSet):
    inode_type_name = {
        1: "list of bad blocks",
        2: "Root directory",
        3: "User quota inode",
        4: "Group quota inode",
        5: "Boot loader",
        6: "Undelete directory",
        7: "Reserved group descriptors",
        8: "EXT3 journal"
    }
    static_size = (68 + 15 * 4) * 8

    def __init__(self, parent, name, index):
        FieldSet.__init__(self, parent, name, None)
        self.uniq_id = 1 + index

    def createDescription(self):
        desc = "Inode %s: " % self.uniq_id
        size = self["size"].value + (self['size_high'].value << 32)
        size = humanFilesize(size)
        if self["links_count"].value == 0:
            desc += "(unused)"
        elif 11 <= self.uniq_id:
            desc += "%s, size=%s, mode=%s" % (self.describe_file(),
                                              size, self['mode'].display)
        elif self.uniq_id in self.inode_type_name:
            desc += self.inode_type_name[self.uniq_id]
            if self.uniq_id == 2:
                desc += ", size=%s, mode=%s" % (size, self['mode'].display)
        else:
            desc += "special"
        return desc

    def describe_file(self):
        filetype = FileMode.file_type_letter.get(
            self['mode/file_type'].value, '-')
        out = self["mode/file_type"].display
        if filetype in 'pds-':
            return out
        elif filetype in 'bc':
            # block/char device
            return out + ' (%d,%d)' % (self['dev_major'].value, self['dev_minor'].value)
        elif filetype == 'l':
            # symlink
            if self['size'].value <= 60:
                return out + ' (-> %s)' % (self['link_target'].value)
        return out

    def is_fast_symlink(self):
        self.seekByte(4 * 15 + 4)
        acl = UInt32(self, "file_acl")

        b = 0
        if acl.value > 0:
            b = (2 << self["/superblock/log_block_size"].value)

        return (self['blocks'].value - b == 0)

    def createFields(self):
        os = self["/superblock/creator_os"].value

        yield FileMode(self, "mode", "File mode")
        yield UInt16(self, "uid", "User ID")
        yield UInt32(self, "size", "File size (in bytes)")
        yield TimestampUnix32(self, "atime", "Last access time")
        yield TimestampUnix32(self, "ctime", "Last inode change time")
        yield TimestampUnix32(self, "mtime", "Last data modification time")
        yield TimestampUnix32(self, "dtime", "Deletion time")
        yield UInt16(self, "gid", "Group ID")
        yield UInt16(self, "links_count", "Hard link count")
        yield UInt32(self, "blocks", "Number of blocks")
        yield InodeFlags(self, "flags", "Flags")
        if os == SuperBlock.OS_LINUX:
            yield UInt32(self, "version_high", "High 32 bits of the version field")
        else:
            yield NullBytes(self, "reserved[]", 4, "Reserved")

        filetype = FileMode.file_type_letter.get(
            self['mode/file_type'].value, '-')
        if filetype in 'bc':
            yield UInt8(self, "dev_minor", "Minor number of the block/char device")
            yield UInt8(self, "dev_major", "Major number of the block/char device")
            yield NullBytes(self, "block_unused", 58)
        elif filetype == 'l' and self.is_fast_symlink():
            yield String(self, "link_target", self['size'].value, "Target filename of this symlink")
            rest = 60 - self['size'].value
            if rest:
                yield NullBytes(self, "block_unused", rest)
        elif self['flags/extents'].value:
            yield ExtentNode(self, "extent_root", size=60 * 8)
        else:
            for index in range(15):
                yield UInt32(self, "block[]")

        yield UInt32(self, "version", "File version, for NFS")
        yield UInt32(self, "file_acl", "File ACL of the xattr block")
        yield UInt32(self, "size_high", "High 32 bits of the file size")
        yield UInt32(self, "faddr", "Block where the fragment of the file resides (obsolete)")

        if os == SuperBlock.OS_LINUX:
            yield UInt16(self, "blocks_high", "High 16 bits of the block count")
            yield UInt16(self, "file_acl_high", "High 16 bits of the xattr block")
            yield UInt16(self, "uid_high", "High 16 bits of user ID")
            yield UInt16(self, "gid_high", "High 16 bits of group ID")
            yield UInt16(self, "checksum", "inode checksum")
            yield NullBytes(self, "reserved[]", 2, "Reserved")
        elif os == SuperBlock.OS_HURD:
            yield UInt8(self, "frag", "Number of fragments in the block")
            yield UInt8(self, "fsize", "Fragment size")
            yield UInt16(self, "mode_high", "High 16 bits of mode")
            yield UInt16(self, "uid_high", "High 16 bits of user ID")
            yield UInt16(self, "gid_high", "High 16 bits of group ID")
            yield UInt32(self, "author", "Author ID (?)")
        else:
            yield RawBytes(self, "raw", 12, "Reserved")


class Directory(Parser):
    PARSER_TAGS = {
        "description": "Directory of EXT2/EXT3 file system",
    }
    endian = LITTLE_ENDIAN

    def createFields(self):
        while self.current_size < self.size:
            yield DirectoryEntry(self, "entry[]")

    def validate(self):
        return True


class Bitmap(FieldSet):
    itemdesc = 'Item'

    def __init__(self, parent, name, start, count, size, description, **kw):
        description = "%s: %s items" % (description, count)
        FieldSet.__init__(self, parent, name, description, size=size, **kw)
        self.start = 1 + start
        self.count = count

    def createFields(self):
        for index in range(self.count):
            yield Bit(self, "item[]", "%s %s" % (self.itemdesc, self.start + index))
        if self.size > self.count:
            yield PaddingBytes(self, "padding", (self.size - self.count) // 8, pattern='\xff')


class BlockBitmap(Bitmap):
    itemdesc = 'Block'


class InodeBitmap(Bitmap):
    itemdesc = 'Inode'


class GroupDescriptorFlags(Flags):
    static_size = 16

    def createFields(self):
        yield Bit(self, "inode_uninit", "inode table and bitmap are not initialized")
        yield Bit(self, "block_uninit", "block bitmap is not initialized")
        yield Bit(self, "inode_zeroed", "inode table is zeroed")
        yield PaddingBits(self, "reserved[]", 13)


class GroupDescriptor(FieldSet):
    static_size = 32 * 8

    def __init__(self, parent, name, index):
        FieldSet.__init__(self, parent, name)
        self.uniq_id = index

    def createDescription(self):
        blocks_per_group = self["/superblock/blocks_per_group"].value
        start = self.uniq_id * blocks_per_group
        end = start + blocks_per_group
        return "Group descriptor: blocks %s-%s" % (start, end)

    def createFields(self):
        yield UInt32(self, "block_bitmap", "Starting block index of the block bitmap")
        yield UInt32(self, "inode_bitmap", "Starting block index of the inode bitmap")
        yield UInt32(self, "inode_table", "Starting block index of the inode table")
        yield UInt16(self, "free_blocks_count", "Number of free blocks")
        yield UInt16(self, "free_inodes_count", "Number of free inodes")
        yield UInt16(self, "used_dirs_count", "Number of inodes allocated to directories")
        yield GroupDescriptorFlags(self, "bg_flags", "Block group flags")
        yield UInt32(self, "exclude_bitmap", "Starting block index of the snapshot exclusion bitmap")
        yield UInt16(self, "block_bitmap_csum", "Block bitmap checksum")
        yield UInt16(self, "inode_bitmap_csum", "Inode bitmap checksum")
        yield UInt16(self, "itable_unused", "Number of unused inodes")
        yield UInt16(self, "checksum", "Group descriptor checksum")


class FeatureCompatFlags(Flags):
    static_size = 32

    def createFields(self):
        yield Bit(self, "dir_prealloc", "Directory preallocation")
        yield Bit(self, "imagic_inodes", "imagic inodes - use unclear")
        yield Bit(self, "has_journal", "Has a journal")
        yield Bit(self, "ext_attr", "Supports extended attributes")
        yield Bit(self, "resize_inode", "Has reserved GDT blocks for FS expansion")
        yield Bit(self, "dir_index", "Has directory indices")
        yield Bit(self, "lazy_bg", "Lazy block groups (not used by Linux)")
        yield Bit(self, "exclude_inode", "Exclude inode (deprecated)")
        yield Bit(self, "exclude_bitmap", "Exclude bitmap (unused)")
        yield Bit(self, "sparse_super2", "Sparse Super Block v2")
        yield PaddingBits(self, "reserved[]", 22)


class FeatureIncompatFlags(Flags):
    static_size = 32

    def createFields(self):
        yield Bit(self, "compression", "Compression")
        yield Bit(self, "filetype", "Directory entries record file type")
        yield Bit(self, "recover", "FS needs recovery")
        yield Bit(self, "journal_dev", "FS has a separate journal device")
        yield Bit(self, "meta_bg", "Meta block groups")
        yield Bit(self, "reserved[]")
        yield Bit(self, "extents", "Files use extents")
        yield Bit(self, "64bit", "FS can have up to 2^64 blocks")
        yield Bit(self, "mmp", "Multiple mount protection")
        yield Bit(self, "flex_bg", "Flexible block groups")
        yield Bit(self, "ea_inode", "Inodes can be used for large xattrs")
        yield Bit(self, "reserved[]")
        yield Bit(self, "dirdata", "Data in directory entry")
        yield Bit(self, "csum_seed", "Metadata checksum seed in the superblock")
        yield Bit(self, "largedir", "Large directory >2GB, or 3-level htree")
        yield Bit(self, "inline_data", "Data in inode")
        yield Bit(self, "encrypt", "Encrypted inodes present")
        yield PaddingBits(self, "reserved[]", 15)


class FeatureROCompatFlags(Flags):
    static_size = 32

    def createFields(self):
        yield Bit(self, "sparse_super", "Sparse superblocks")
        yield Bit(self, "large_file", "The FS has been used to store a file >2GiB")
        yield Bit(self, "btree_dir")
        yield Bit(self, "huge_file", "The FS has files whose sizes are in units of logical blocks")
        yield Bit(self, "gdt_csum", "Group descriptors have checksums")
        yield Bit(self, "dir_nlink", "ext3 32k subdir limit does not apply")
        yield Bit(self, "extra_isize", "FS has large inodes")
        yield Bit(self, "has_snapshot", "FS has a snapshot")
        yield Bit(self, "quota", "FS has quotas")
        yield Bit(self, "bigalloc", "FS supports allocating extents in units of clusters (fragments)")
        yield Bit(self, "metadata_csum", "FS supports metadata checksums")
        yield Bit(self, "replica", "FS supports replicas")
        yield Bit(self, "readonly", "FS is readonly")
        yield PaddingBits(self, "reserved[]", 19)


class DefaultMountOptionFlags(Flags):
    static_size = 32

    def createFields(self):
        yield Bit(self, "debug", "Print debug info upon mount")
        yield Bit(self, "bsdgroups", "New files take the gid of the containing dir")
        yield Bit(self, "xattr_user", "Support userspace-provided xattrs")
        yield Bit(self, "acl", "Support POSIX access control lists")
        yield Bit(self, "uid16", "Do not support 32-bit UIDs")
        yield Enum(Bits(self, "jmode", 2, "Journaling mode"),
                   {0: "none", 1: "data", 2: "ordered", 3: "wback"})
        yield Bit(self, "reserved[]")
        yield Bit(self, "nobarrier", "Disable write flushes")
        yield Bit(self, "block_validity", "Track metadata blocks to avoid treating them as data blocks")
        yield Bit(self, "discard", "Tell the storage device when blocks become unused")
        yield Bit(self, "nodelalloc", "Disable delayed allocation")
        yield PaddingBits(self, "reserved[]", 20)

    def createDisplay(self):
        out = []
        for field in self:
            if field.name == 'jmode':
                if field.value:
                    out.append('jmode_' + field.display)
            elif field.value:
                out.append(field.name)
        return '{' + ', '.join(out) + '}'


class SuperBlock(FieldSet):
    static_size = 1024 * 8

    OS_LINUX = 0
    OS_HURD = 1
    os_name = {
        0: "Linux",
        1: "Hurd",
        2: "Masix",
        3: "FreeBSD",
        4: "Lites",
        5: "WinNT"
    }
    state_desc = {
        1: "Valid (Unmounted cleanly)",
        2: "Error (Errors detected)",
        4: "Orphan FS (Orphans being recovered)",
    }
    error_handling_desc = {1: "Continue", 2: "Remount R/O", 3: "Panic"}
    revision_levels = {
        0: "The good old (original) format",
        1: "V2 format w/ dynamic inode sizes",
    }
    htree_hash_algo_desc = {
        0: 'Legacy', 1: 'Half MD4', 2: 'TEA', 3: 'Legacy unsigned', 4: 'Half MD4 unsigned', 5: 'TEA unsigned'
    }

    def __init__(self, parent, name):
        FieldSet.__init__(self, parent, name)
        self._group_count = None

    def createDescription(self):
        if self["feature_compat/has_journal"].value:
            fstype = "ext3"
        else:
            fstype = "ext2"
        return "Superblock: %s file system" % fstype

    def createFields(self):
        yield UInt32(self, "inodes_count", "Total inode count")
        yield UInt32(self, "blocks_count", "Total block count")
        yield UInt32(self, "r_blocks_count", "Reserved (superuser-only) block count")
        yield UInt32(self, "free_blocks_count", "Free block count")
        yield UInt32(self, "free_inodes_count", "Free inode count")
        yield UInt32(self, "first_data_block", "First data block")
        yield UInt32(self, "log_block_size", "Block size = 2**(10+log_block_size)")
        yield UInt32(self, "log_frag_size", "Cluster size = 2**log_frag_size")
        yield UInt32(self, "blocks_per_group", "Blocks per group")
        yield UInt32(self, "frags_per_group", "Fragments per group")
        yield UInt32(self, "inodes_per_group", "Inodes per group")
        yield TimestampUnix32(self, "mtime", "Mount time")
        yield TimestampUnix32(self, "wtime", "Write time")
        yield UInt16(self, "mnt_count", "Mount count since the last fsck")
        yield UInt16(self, "max_mnt_count", "Max mount count before fsck is needed")
        yield UInt16(self, "magic", "Magic number (0xEF53)")
        yield Enum(UInt16(self, "state", "File system state"), self.state_desc)
        yield Enum(UInt16(self, "errors", "Behaviour when detecting errors"), self.error_handling_desc)
        yield UInt16(self, "minor_rev_level", "Minor revision level")
        yield TimestampUnix32(self, "last_check", "Time of last check")
        yield textHandler(UInt32(self, "check_interval", "Maximum time between checks"), self.postMaxTime)
        yield Enum(UInt32(self, "creator_os", "Creator OS"), self.os_name)
        yield Enum(UInt32(self, "rev_level", "Revision level"), self.revision_levels)
        yield UInt16(self, "def_resuid", "Default uid for reserved blocks")
        yield UInt16(self, "def_resgid", "Default gid for reserved blocks")
        yield UInt32(self, "first_ino", "First non-reserved inode")
        yield UInt16(self, "inode_size", "Size of inode structure")
        yield UInt16(self, "block_group_nr", "Block group # of this superblock")
        yield FeatureCompatFlags(self, "feature_compat", "Compatible feature set (can mount even if these features are unsupported)")
        yield FeatureIncompatFlags(self, "feature_incompat", "Incompatible feature set (must support all features to mount)")
        yield FeatureROCompatFlags(self, "feature_ro_compat", "Read-only compatible feature set (can only mount r/o if a feature is unsupported)")
        yield UUID(self, "uuid", "128-bit UUID for volume")
        yield String(self, "volume_name", 16, "Volume name", strip="\0")
        yield String(self, "last_mounted", 64, "Directory where last mounted", strip="\0")
        yield UInt32(self, "compression", "For compression (algorithm usage bitmap)")
        yield UInt8(self, "prealloc_blocks", "Number of blocks to try to preallocate")
        yield UInt8(self, "prealloc_dir_blocks", "Number to preallocate for directories")
        yield UInt16(self, "reserved_gdt_blocks", "Number of reserved GDT entries for future expansion")
        yield RawBytes(self, "journal_uuid", 16, "UUID of journal superblock")
        yield UInt32(self, "journal_inum", "Inode number of journal file")
        yield UInt32(self, "journal_dev", "Device number of journal file (if ext_journal feature is set)")
        yield UInt32(self, "last_orphan", "Start of list of orphaned inodes to delete")
        # ext3 stuff
        yield RawBytes(self, "hash_seed", 16, "Seeds used for the directory indexing hash algorithm")
        yield Enum(UInt8(self, "def_hash_version", "Default hash version for directory indexing"),
                   self.htree_hash_algo_desc)
        yield UInt8(self, "jnl_backup_type", "Does jnl_blocks contain a backup of i_block and i_size?")
        yield UInt16(self, "desc_size", "Size of group descriptors (if 64bit feature is set)")
        yield DefaultMountOptionFlags(self, "default_mount_opts", "Default mount options")
        yield UInt32(self, "first_meta_bg", "First metablock block group (if meta_bg feature is set)")
        yield TimestampUnix32(self, "mkfs_time", "When the filesystem was created")
        yield RawBytes(self, "jnl_blocks", 17 * 4, "Backup of the journal inode's i_block and i_size")

        yield PaddingBytes(self, "reserved[]", (1024 << self['log_block_size'].value) - self.current_size // 8)

    def _getGroupCount(self):
        if self._group_count is None:
            # Calculate number of groups
            blocks_per_group = self["blocks_per_group"].value
            self._group_count = (self["blocks_count"].value - self[
                                 "first_data_block"].value + (blocks_per_group - 1)) // blocks_per_group
        return self._group_count
    group_count = property(_getGroupCount)

    def postMaxTime(self, chunk):
        return humanDuration(chunk.value * 1000)


class GroupDescriptors(FieldSet):

    def __init__(self, parent, name, count):
        FieldSet.__init__(self, parent, name)
        self.count = count

    def createDescription(self):
        return "Group descriptors: %s items" % self.count

    def createFields(self):
        for index in range(self.count):
            yield GroupDescriptor(self, "group[]", index)


class InodeTable(FieldSet):

    def __init__(self, parent, name, start, count):
        FieldSet.__init__(self, parent, name)
        self.start = start
        self.count = count
        inode_size = self["/superblock/inode_size"].value
        if inode_size == 0:
            inode_size = 128
        self._size = self.count * inode_size * 8

    def createDescription(self):
        return "Inodes: %s items" % self.count

    def createFields(self):
        for index in range(self.start, self.start + self.count):
            yield Inode(self, "inode[]", index)


class IndirectBlock(FieldSet):

    def __init__(self, parent, name, size):
        FieldSet.__init__(self, parent, name, size=size * 8)
        self.count = size // 4

    def createFields(self):
        for i in range(self.count):
            yield UInt32(self, "block[]")


class Group(SeekableFieldSet):

    def __init__(self, parent, name, descriptor):
        self.superblock = parent['superblock']
        block_size = 1024 << self.superblock['log_block_size'].value
        size = block_size * self.superblock['blocks_per_group'].value * 8
        SeekableFieldSet.__init__(self, parent, name, size=size)
        self.descriptor = descriptor

    def createDescription(self):
        desc = "Group %s" % (self.descriptor.uniq_id)
        return desc

    def seekBlock(self, block):
        self.seekBit(block * self.root.block_size * 8, relative=False)

    def createFields(self):
        superblock = self["/superblock"]

        # Compute number of block and inodes
        block_count = superblock["blocks_per_group"].value
        inode_count = superblock["inodes_per_group"].value
        block_index = self.descriptor.uniq_id * block_count
        inode_index = self.descriptor.uniq_id * inode_count
        if (block_count % 8) != 0:
            raise ParserError("Invalid block count")
        if (inode_count % 8) != 0:
            raise ParserError("Invalid inode count")
        block_count = min(block_count, superblock[
                          "blocks_count"].value - block_index)
        inode_count = min(inode_count, superblock[
                          "inodes_count"].value - inode_index)

        bitmap_block_size = self.root.block_size * 8
        block_bitmap_size = (block_count + bitmap_block_size -
                             1) // bitmap_block_size * bitmap_block_size
        inode_bitmap_size = (inode_count + bitmap_block_size -
                             1) // bitmap_block_size * bitmap_block_size

        self.seekBlock(self.descriptor["block_bitmap"].value)
        yield BlockBitmap(self, "block_bitmap", block_index, block_count, block_bitmap_size, "Block bitmap")

        self.seekBlock(self.descriptor["inode_bitmap"].value)
        yield InodeBitmap(self, "inode_bitmap", inode_index, inode_count, inode_bitmap_size, "Inode bitmap")

        self.seekBlock(self.descriptor["inode_table"].value)
        yield InodeTable(self, "inode_table", inode_index, inode_count)

        inode_bitmap = self['inode_bitmap']
        for i, inode in enumerate(self['inode_table'].array('inode')):
            if not inode_bitmap['item[%d]' % i].value:
                continue
            if inode['blocks'].value == 0:
                continue
            blocks = inode.array('block')
            if not blocks:
                continue
            if inode['mode/file_type'].display == 'Directory':
                parser = Directory
            else:
                parser = None
            group = FragmentGroup(parser=parser)
            for b in range(12):
                if not blocks[b].value:
                    continue
                self.seekBlock(blocks[b].value)
                yield CustomFragment(self, "inode[%d]block[]" % i, self.root.block_size * 8,
                                     None, group=group)
            if blocks[12].value:
                # indirect block
                self.seekBlock(blocks[12].value)
                indirect = IndirectBlock(self, "inode[%d]indirect" % i, self.root.block_size)
                yield indirect
                for b in indirect.array('block'):
                    if not b.value:
                        continue
                    self.seekBlock(b.value)
                    yield CustomFragment(self, "inode[%d]block[]" % i, self.root.block_size * 8,
                                         None, group=group)
            if blocks[13].value:
                # TODO: double-indirect block
                pass
            if blocks[14].value:
                # TODO: triple-indirect block
                pass


class EXT2_FS(HachoirParser, RootSeekableFieldSet):
    """
    Parse an EXT2 or EXT3 partition.

    Attributes:
       * block_size: Size of a block (in bytes)

    Fields:
       * superblock: Most important block, store most important informations
       * ...
    """
    PARSER_TAGS = {
        "id": "ext2",
        "category": "file_system",
        "description": "EXT2/EXT3 file system",
        "min_size": (1024 * 2) * 8,
        "magic": (
            # (magic, state=valid)
            (b"\x53\xEF\1\0", 1080 * 8),
            # (magic, state=error)
            (b"\x53\xEF\2\0", 1080 * 8),
            # (magic, state=error)
            (b"\x53\xEF\4\0", 1080 * 8),
        ),
    }
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes((1024 + 56) * 8, 2) != b"\x53\xEF":
            return "Invalid magic number"
        if not (0 <= self["superblock/log_block_size"].value <= 2):
            return "Invalid (log) block size"
        if self["superblock/inode_size"].value not in (0, 128):
            return "Unsupported inode size"
        return True

    def createFields(self):
        # Skip something (what is stored here? MBR?)
        yield NullBytes(self, "padding[]", 1024)

        # Read superblock
        superblock = SuperBlock(self, "superblock")
        yield superblock
        self.block_size = 1024 << superblock["log_block_size"].value  # in bytes

        if self.block_size == 1024:
            self.seekBlock(2)
        else:
            # block_size > 1024
            self.seekBlock(1)

        # Read groups' descriptor
        groups = GroupDescriptors(self, "group_desc", superblock.group_count)
        yield groups

        for group in groups.array('group'):
            self.seekBlock(min(group[f].value for f in [
                           'block_bitmap', 'inode_bitmap', 'inode_table']))
            yield Group(self, "group[]", group)

    def seekBlock(self, block):
        self.seekBit(block * self.block_size * 8)

    def getSuperblock(self):
        # FIXME: Use superblock copy if main superblock is invalid
        return self["superblock"]

    def createDescription(self):
        superblock = self.getSuperblock()
        block_size = 1024 << superblock["log_block_size"].value
        nb_block = superblock["blocks_count"].value
        total = nb_block * block_size
        used = (superblock["free_blocks_count"].value) * block_size
        desc = "EXT2/EXT3"
        if "group[0]/inode_table/inode[7]/blocks" in self:
            if 0 < self["group[0]/inode_table/inode[7]/blocks"].value:
                desc = "EXT3"
            else:
                desc = "EXT2"
        return desc + " file system: total=%s, used=%s, block=%s" % (
            humanFilesize(total), humanFilesize(used),
            humanFilesize(block_size))



from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN, MIDDLE_ENDIAN
from hachoir.field import GenericFieldSet
from hachoir.core.log import Logger
import hachoir.core.config as config


class Parser(GenericFieldSet):
    """
    A parser is the root of all other fields. It create first level of fields
    and have special attributes and methods:
    - endian: Byte order (L{BIG_ENDIAN}, L{LITTLE_ENDIAN} or L{MIDDLE_ENDIAN}) of input data ;
    - stream: Data input stream (set in L{__init__()}) ;
    - size: Field set size will be size of input stream.
    """

    def __init__(self, stream, description=None):
        """
        Parser constructor

        @param stream: Data input stream (see L{InputStream})
        @param description: (optional) String description
        """
        # Check arguments
        assert hasattr(self, "endian") \
            and self.endian in (BIG_ENDIAN, LITTLE_ENDIAN, MIDDLE_ENDIAN)

        # Call parent constructor
        GenericFieldSet.__init__(
            self, None, "root", stream, description, stream.askSize(self))

    def close(self):
        self.stream.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _logger(self):
        return Logger._logger(self)

    def _setSize(self, size):
        self._truncate(size)
        self.raiseEvent("field-resized", self)
    size = property(lambda self: self._size, doc="Size in bits")

    path = property(lambda self: "/")

    # dummy definition to prevent hachoir-core from depending on hachoir-parser
    autofix = property(lambda self: config.autofix)



from hachoir.core.dict import UniqKeyError
from hachoir.field import MissingField, Float32, Float64, FakeArray
from hachoir.editor import createEditableField, EditorError
from collections import deque  # Python 2.4
import weakref  # Python 2.1
import struct


class EditableFieldSet(object):
    MAX_SIZE = (1 << 40)  # Arbitrary limit to catch errors
    is_field_set = True

    def __init__(self, parent, fieldset):
        self._parent = parent
        self.input = fieldset  # original FieldSet
        self._fields = {}      # cache of editable fields
        self._deleted = set()  # Names of deleted fields
        self._inserted = {}    # Inserted field (name => list of field,
        # where name is the name after)

    def array(self, key):
        # FIXME: Use cache?
        return FakeArray(self, key)

    def _getParent(self):
        return self._parent
    parent = property(_getParent)

    def _isAltered(self):
        if self._inserted:
            return True
        if self._deleted:
            return True
        return any(field.is_altered for field in self._fields.values())
    is_altered = property(_isAltered)

    def reset(self):
        """
        Reset the field set and the input field set.
        """
        for key, field in self._fields.items():
            if not field.is_altered:
                del self._fields[key]
        self.input.reset()

    def __len__(self):
        return len(self.input) \
            - len(self._deleted) \
            + sum(len(new) for new in self._inserted.values())

    def __iter__(self):
        for field in self.input:
            name = field.name
            if name in self._inserted:
                for newfield in self._inserted[name]:
                    yield weakref.proxy(newfield)
            if name not in self._deleted:
                yield self[name]
        if None in self._inserted:
            for newfield in self._inserted[None]:
                yield weakref.proxy(newfield)

    def insertBefore(self, name, *new_fields):
        self._insert(name, new_fields, False)

    def insertAfter(self, name, *new_fields):
        self._insert(name, new_fields, True)

    def insert(self, *new_fields):
        self._insert(None, new_fields, True)

    def _insert(self, key, new_fields, next):
        """
        key is the name of the field before which new_fields
        will be inserted. If next is True, the fields will be inserted
        _after_ this field.
        """
        # Set unique field name
        for field in new_fields:
            if field._name.endswith("[]"):
                self.input.setUniqueFieldName(field)

        # Check that there is no duplicate in inserted fields
        new_names = list(field.name for field in new_fields)
        names_set = set(new_names)
        if len(names_set) != len(new_fields):
            duplicates = (name for name in names_set if 1 <
                          new_names.count(name))
            raise UniqKeyError(
                "Duplicates in inserted fields: %s" % ", ".join(duplicates))

        # Check that field names are not in input
        if self.input:  # Write special version for NewFieldSet?
            for name in new_names:
                if name in self.input and name not in self._deleted:
                    raise UniqKeyError("Field name '%s' already exists" % name)

        # Check that field names are not in inserted fields
        for fields in self._inserted.values():
            for field in fields:
                if field.name in new_names:
                    raise UniqKeyError(
                        "Field name '%s' already exists" % field.name)

        # Input have already inserted field?
        if key in self._inserted:
            if next:
                self._inserted[key].extend(reversed(new_fields))
            else:
                self._inserted[key].extendleft(reversed(new_fields))
            return

        # Whould like to insert in inserted fields?
        if key:
            for fields in self._inserted.values():
                names = [item.name for item in fields]
                try:
                    pos = names.index(key)
                except ValueError:
                    continue
                if 0 <= pos:
                    if next:
                        pos += 1
                    fields.rotate(-pos)
                    fields.extendleft(reversed(new_fields))
                    fields.rotate(pos)
                    return

            # Get next field. Use None if we are at the end.
            if next:
                index = self.input[key].index + 1
                try:
                    key = self.input[index].name
                except IndexError:
                    key = None

            # Check that field names are not in input
            if key not in self.input:
                raise MissingField(self, key)

        # Insert in original input
        self._inserted[key] = deque(new_fields)

    def _getDescription(self):
        return self.input.description
    description = property(_getDescription)

    def _getStream(self):
        # FIXME: This property is maybe a bad idea since address may be
        # differents
        return self.input.stream
    stream = property(_getStream)

    def _getName(self):
        return self.input.name
    name = property(_getName)

    def _getEndian(self):
        return self.input.endian
    endian = property(_getEndian)

    def _getAddress(self):
        if self._parent:
            return self._parent._getFieldAddress(self.name)
        else:
            return 0
    address = property(_getAddress)

    def _getAbsoluteAddress(self):
        address = self.address
        current = self._parent
        while current:
            address += current.address
            current = current._parent
        return address
    absolute_address = property(_getAbsoluteAddress)

    def hasValue(self):
        return False
#        return self._parent.input[self.name].hasValue()

    def _getSize(self):
        if self.is_altered:
            return sum(field.size for field in self)
        else:
            return self.input.size
    size = property(_getSize)

    def _getPath(self):
        return self.input.path
    path = property(_getPath)

    def _getOriginalField(self, name):
        assert name in self.input
        return self.input[name]

    def _getFieldInputAddress(self, name):
        """
        Absolute address of a field from the input field set.
        """
        assert name in self.input
        return self.input[name].absolute_address

    def _getFieldAddress(self, name):
        """
        Compute relative address of a field. The operation takes care of
        deleted and resized fields.
        """
        # assert name not in self._deleted
        addr = 0
        for field in self:
            if field.name == name:
                return addr
            addr += field.size
        raise MissingField(self, name)

    def _getItemByPath(self, path):
        if not path[0]:
            path = path[1:]
        field = self
        for name in path:
            field = field[name]
        return field

    def __contains__(self, name):
        try:
            field = self[name]
            return (field is not None)
        except MissingField:
            return False

    def __getitem__(self, key):
        """
        Create a weak reference to an editable field (EditableField) for the
        field with specified name. If the field is removed later, using the
        editable field will raise a weakref.ReferenceError exception.

        May raise a MissingField error if the field doesn't exist in original
        field set or it has been deleted.
        """
        if "/" in key:
            return self._getItemByPath(key.split("/"))
        if isinstance(key, int):
            raise EditorError("Integer index are not supported")

        if (key in self._deleted) or (key not in self.input):
            raise MissingField(self, key)
        if key not in self._fields:
            field = self.input[key]
            if field.is_field_set:
                self._fields[key] = createEditableFieldSet(self, field)
            else:
                self._fields[key] = createEditableField(self, field)
        return weakref.proxy(self._fields[key])

    def __delitem__(self, name):
        """
        Remove a field from the field set. May raise an MissingField exception
        if the field has already been deleted.
        """
        parts = name.partition('/')
        if parts[2]:
            fieldset = self[parts[0]]
            del fieldset[parts[2]]
            return
        if name in self._deleted:
            raise MissingField(self, name)
        self._deleted.add(name)
        if name in self._fields:
            del self._fields[name]

    def writeInto(self, output):
        """
        Write the content if this field set into the output stream
        (OutputStream).
        """
        if not self.is_altered:
            # Not altered: just copy bits/bytes
            input = self.input
            if input.size % 8:
                output.copyBitsFrom(input.stream,
                                    input.absolute_address, input.size, input.endian)
            else:
                output.copyBytesFrom(input.stream,
                                     input.absolute_address, input.size // 8)
        else:
            # Altered: call writeInto() method of each field
            realaddr = 0
            for field in self:
                field.writeInto(output)
                realaddr += field.size

    def _getValue(self):
        raise EditorError('Field set "%s" has no value' % self.path)

    def _setValue(self, value):
        raise EditorError('Field set "%s" value is read only' % self.path)
    value = property(_getValue, _setValue, "Value of field")


class EditableFloat(EditableFieldSet):
    _value = None

    def _isAltered(self):
        return (self._value is not None)
    is_altered = property(_isAltered)

    def writeInto(self, output):
        if self._value is not None:
            self._write(output)
        else:
            EditableFieldSet.writeInto(self, output)

    def _write(self, output):
        format = self.input.struct_format
        raw = struct.pack(format, self._value)
        output.writeBytes(raw)

    def _setValue(self, value):
        self.parent._is_altered = True
        self._value = value
    value = property(EditableFieldSet._getValue, _setValue)


def createEditableFieldSet(parent, field):
    cls = field.__class__
    # FIXME: Support Float80
    if cls in (Float32, Float64):
        return EditableFloat(parent, field)
    else:
        return EditableFieldSet(parent, field)


class NewFieldSet(EditableFieldSet):

    def __init__(self, parent, name):
        EditableFieldSet.__init__(self, parent, None)
        self._name = name
        self._endian = parent.endian

    def __iter__(self):
        if None in self._inserted:
            return iter(self._inserted[None])
        else:
            raise StopIteration()

    def _getName(self):
        return self._name
    name = property(_getName)

    def _getEndian(self):
        return self._endian
    endian = property(_getEndian)

    is_altered = property(lambda self: True)


def createEditor(fieldset):
    return EditableFieldSet(None, fieldset)



"""
GZIP archive parser.

Author: Victor Stinner
"""

from hachoir.parser import Parser
from hachoir.field import (
    UInt8, UInt16, UInt32, Enum, TimestampUnix32,
    Bit, CString, SubFile,
    NullBits, Bytes, RawBytes)
from hachoir.core.text_handler import textHandler, hexadecimal, filesizeHandler
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.common.deflate import Deflate


class GzipParser(Parser):
    endian = LITTLE_ENDIAN
    PARSER_TAGS = {
        "id": "gzip",
        "category": "archive",
        "file_ext": ("gz",),
        "mime": ("application/x-gzip",),
        "min_size": 18 * 8,
        "magic_regex": (
            # (magic, compression=deflate, <flags>, <mtime>, )
            (b'\x1F\x8B\x08.{5}[\0\2\4\6][\x00-\x0D]', 0),
        ),
        "description": "gzip archive",
    }
    os_name = {
        0: "FAT filesystem",
        1: "Amiga",
        2: "VMS (or OpenVMS)",
        3: "Unix",
        4: "VM/CMS",
        5: "Atari TOS",
        6: "HPFS filesystem (OS/2, NT)",
        7: "Macintosh",
        8: "Z-System",
        9: "CP/M",
        10: "TOPS-20",
        11: "NTFS filesystem (NT)",
        12: "QDOS",
        13: "Acorn RISCOS",
    }
    COMPRESSION_NAME = {
        8: "deflate",
    }

    def validate(self):
        if self["signature"].value != b'\x1F\x8B':
            return "Invalid signature"
        if self["compression"].value not in self.COMPRESSION_NAME:
            return "Unknown compression method (%u)" % self["compression"].value
        if self["reserved[0]"].value != 0:
            return "Invalid reserved[0] value"
        if self["reserved[1]"].value != 0:
            return "Invalid reserved[1] value"
        if self["reserved[2]"].value != 0:
            return "Invalid reserved[2] value"
        return True

    def createFields(self):
        # Gzip header
        yield Bytes(self, "signature", 2, r"GZip file signature (\x1F\x8B)")
        yield Enum(UInt8(self, "compression", "Compression method"), self.COMPRESSION_NAME)

        # Flags
        yield Bit(self, "is_text", "File content is probably ASCII text")
        yield Bit(self, "has_crc16", "Header CRC16")
        yield Bit(self, "has_extra", "Extra informations (variable size)")
        yield Bit(self, "has_filename", "Contains filename?")
        yield Bit(self, "has_comment", "Contains comment?")
        yield NullBits(self, "reserved[]", 3)
        yield TimestampUnix32(self, "mtime", "Modification time")

        # Extra flags
        yield NullBits(self, "reserved[]", 1)
        yield Bit(self, "slowest", "Compressor used maximum compression (slowest)")
        yield Bit(self, "fastest", "Compressor used the fastest compression")
        yield NullBits(self, "reserved[]", 5)
        yield Enum(UInt8(self, "os", "Operating system"), self.os_name)

        # Optional fields
        if self["has_extra"].value:
            yield UInt16(self, "extra_length", "Extra length")
            yield RawBytes(self, "extra", self["extra_length"].value, "Extra")
        if self["has_filename"].value:
            yield CString(self, "filename", "Filename", charset="ISO-8859-1")
        if self["has_comment"].value:
            yield CString(self, "comment", "Comment")
        if self["has_crc16"].value:
            yield textHandler(UInt16(self, "hdr_crc16", "CRC16 of the header"),
                              hexadecimal)

        if self._size is None:  # TODO: is it possible to handle piped input?
            raise NotImplementedError()

        # Read file
        size = (self._size - self.current_size) // 8 - 8  # -8: crc32+size
        if 0 < size:
            if self["has_filename"].value:
                filename = self["filename"].value
            else:
                for tag, filename in self.stream.tags:
                    if tag == "filename" and filename.endswith(".gz"):
                        filename = filename[:-3]
                        break
                else:
                    filename = None
            yield Deflate(SubFile(self, "file", size, filename=filename))

        # Footer
        yield textHandler(UInt32(self, "crc32",
                                 "Uncompressed data content CRC32"), hexadecimal)
        yield filesizeHandler(UInt32(self, "size", "Uncompressed size"))

    def createDescription(self):
        desc = "gzip archive"
        info = []
        if "filename" in self:
            info.append('filename "%s"' % self["filename"].value)
        if "size" in self:
            info.append("was %s" % self["size"].display)
        if self["mtime"].value:
            info.append(self["mtime"].display)
        return "%s: %s" % (desc, ", ".join(info))



from hachoir.core.endian import endian_name
from hachoir.core.tools import makeUnicode
from hachoir.core.dict import Dict
from hachoir.core.error import error
from hachoir.core.log import Logger
from hachoir.metadata.metadata_item import (
    MIN_PRIORITY, MAX_PRIORITY, QUALITY_NORMAL)
from hachoir.metadata.register import registerAllItems

extractors = {}


class Metadata(Logger):
    header = "Metadata"

    def __init__(self, parent, quality=QUALITY_NORMAL):
        assert isinstance(self.header, str)

        # Limit to 0.0 .. 1.0
        if parent:
            quality = parent.quality
        else:
            quality = min(max(0.0, quality), 1.0)

        object.__init__(self)
        object.__setattr__(self, "_Metadata__data", {})
        object.__setattr__(self, "quality", quality)
        header = self.__class__.header
        object.__setattr__(self, "_Metadata__header", header)

        registerAllItems(self)

    def _logger(self):
        pass

    def __setattr__(self, key, value):
        """
        Add a new value to data with name 'key'. Skip duplicates.
        """
        # Invalid key?
        if key not in self.__data:
            raise KeyError("%s has no metadata '%s'" %
                           (self.__class__.__name__, key))

        # Skip duplicates
        self.__data[key].add(value)

    def setHeader(self, text):
        object.__setattr__(self, "header", text)

    def getItems(self, key):
        try:
            return self.__data[key]
        except LookupError:
            raise ValueError("Metadata has no value '%s'" % key)

    def getItem(self, key, index):
        try:
            return self.getItems(key)[index]
        except (LookupError, ValueError):
            return None

    def has(self, key):
        return 1 <= len(self.getItems(key))

    def get(self, key, default=None, index=0):
        """
        Read first value of tag with name 'key'.

        >>> from datetime import timedelta
        >>> a = RootMetadata()
        >>> a.duration = timedelta(seconds=2300)
        >>> a.get('duration') == timedelta(seconds=2300)
        True
        >>> a.get('author', 'Anonymous')
        'Anonymous'
        """
        item = self.getItem(key, index)
        if item is None:
            if default is None:
                raise ValueError(
                    "Metadata has no value '%s' (index %s)" % (key, index))
            else:
                return default
        return item.value

    def getValues(self, key):
        try:
            data = self.__data[key]
        except LookupError:
            raise ValueError("Metadata has no value '%s'" % key)
        return [item.value for item in data]

    def getText(self, key, default=None, index=0):
        """
        Read first value, as unicode string, of tag with name 'key'.

        >>> from datetime import timedelta
        >>> a = RootMetadata()
        >>> a.duration = timedelta(seconds=2300)
        >>> a.getText('duration')
        '38 min 20 sec'
        >>> a.getText('titre', 'Unknown')
        'Unknown'
        """
        item = self.getItem(key, index)
        if item is not None:
            return item.text
        else:
            return default

    def register(self, data):
        assert data.key not in self.__data
        data.metadata = self
        self.__data[data.key] = data

    def __iter__(self):
        return iter(self.__data.values())

    def __str__(self):
        r"""
        Create a multi-line Unicode string (end of line is "\n") which
        represents all datas.

        >>> a = RootMetadata()
        >>> a.copyright = "© Hachoir"
        >>> print(repr(str(a)))
        'Metadata:\n- Copyright: \xa9 Hachoir'

        @see __str__() and exportPlaintext()
        """
        return "\n".join(self.exportPlaintext())

    def exportPlaintext(self, priority=None, human=True, line_prefix="- ", title=None):
        r"""
        Convert metadata to multi-line Unicode string and skip datas
        with priority lower than specified priority.

        Default priority is Metadata.MAX_PRIORITY. If human flag is True, data
        key are translated to better human name (eg. "bit_rate" becomes
        "Bit rate") which may be translated using gettext.

        If priority is too small, metadata are empty and so None is returned.

        >>> print(RootMetadata().exportPlaintext())
        None
        >>> meta = RootMetadata()
        >>> meta.copyright = "© Hachoir"
        >>> print(repr(meta.exportPlaintext()))
        ['Metadata:', '- Copyright: \xa9 Hachoir']

        @see __str__() and __unicode__()
        """
        if priority is not None:
            priority = max(priority, MIN_PRIORITY)
            priority = min(priority, MAX_PRIORITY)
        else:
            priority = MAX_PRIORITY
        if not title:
            title = self.header
        text = ["%s:" % title]
        for data in sorted(self, key=lambda data: data.priority):
            if priority < data.priority:
                break
            if not data.values:
                continue
            if human:
                title = data.description
            else:
                title = data.key
            for item in data.values:
                if human:
                    value = item.text
                else:
                    value = makeUnicode(item.value)
                text.append("%s%s: %s" % (line_prefix, title, value))
        if 1 < len(text):
            return text
        else:
            return None

    def exportDictionary(self, priority=None, human=True, title=None):
        r"""
        Convert metadata to python Dictionary and skip datas
        with priority lower than specified priority.

        Default priority is Metadata.MAX_PRIORITY. If human flag is True, data
        key are translated to better human name (eg. "bit_rate" becomes
        "Bit rate") which may be translated using gettext.

        If priority is too small, metadata are empty and so None is returned.

        """
        if priority is not None:
            priority = max(priority, MIN_PRIORITY)
            priority = min(priority, MAX_PRIORITY)
        else:
            priority = MAX_PRIORITY
        if not title:
            title = self.header
        text = {}
        text[title] = {}
        for data in sorted(self):
            if priority < data.priority:
                break
            if not data.values:
                continue
            if human:
                field = data.description
            else:
                field = data.key
            text[title][field] = {}
            for item in data.values:
                if human:
                    value = item.text
                else:
                    value = makeUnicode(item.value)
                text[title][field] = value
        return text

    def __bool__(self):
        return any(item for item in self.__data.values())


class RootMetadata(Metadata):

    def __init__(self, quality=QUALITY_NORMAL):
        Metadata.__init__(self, None, quality)


class MultipleMetadata(RootMetadata):
    header = "Common"

    def __init__(self, quality=QUALITY_NORMAL):
        RootMetadata.__init__(self, quality)
        object.__setattr__(self, "_MultipleMetadata__groups", Dict())
        object.__setattr__(self, "_MultipleMetadata__key_counter", {})

    def __contains__(self, key):
        return key in self.__groups

    def __getitem__(self, key):
        return self.__groups[key]

    def iterGroups(self):
        return iter(self.__groups.values)

    def __bool__(self):
        if RootMetadata.__bool__(self):
            return True
        return any(bool(group) for group in self.__groups)

    def addGroup(self, key, metadata, header=None):
        """
        Add a new group (metadata of a sub-document).

        Returns False if the group is skipped, True if it has been added.
        """
        if not metadata:
            self.warning("Skip empty group %s" % key)
            return False
        if key.endswith("[]"):
            key = key[:-2]
            if key in self.__key_counter:
                self.__key_counter[key] += 1
            else:
                self.__key_counter[key] = 1
            key += "[%u]" % self.__key_counter[key]
        if header:
            metadata.setHeader(header)
        self.__groups.append(key, metadata)
        return True

    def exportPlaintext(self, priority=None, human=True, line_prefix="- "):
        common = Metadata.exportPlaintext(self, priority, human, line_prefix)
        if common:
            text = common
        else:
            text = []
        for key, metadata in self.__groups.items():
            if not human:
                title = key
            else:
                title = None
            value = metadata.exportPlaintext(
                priority, human, line_prefix, title=title)
            if value:
                text.extend(value)
        if len(text):
            return text
        else:
            return None

    def exportDictionary(self, priority=None, human=True):
        common = Metadata.exportDictionary(self, priority, human)
        if common:
            text = common
        else:
            text = {}
        for key, metadata in self.__groups.items():
            if not human:
                title = key
            else:
                title = None
            value = metadata.exportDictionary(priority, human, title=title)
            if value:
                text.update(value)
        return text


def registerExtractor(parser, extractor):
    assert parser not in extractors
    assert issubclass(extractor, RootMetadata)
    extractors[parser] = extractor


def extractMetadata(parser, quality=QUALITY_NORMAL):
    """
    Create a Metadata class from a parser. Returns None if no metadata
    extractor does exist for the parser class.
    """
    try:
        extractor = extractors[parser.__class__]
    except KeyError:
        return None
    metadata = extractor(quality)
    try:
        metadata.extract(parser)
    except Exception as err:
        error("Error during metadata extraction: %s" % str(err))
    if metadata:
        metadata.mime_type = parser.mime_type
        metadata.endian = endian_name[parser.endian]
    return metadata



"""
X11 Portable Compiled Font (pcf) parser.

Documents:
- Format for X11 pcf bitmap font files
  http://fontforge.sourceforge.net/pcf-format.html
  (file is based on the X11 sources)

Author: Victor Stinner
Creation date: 2007-03-20
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, Enum,
                           UInt8, UInt32, Bytes, RawBytes, NullBytes,
                           Bit, Bits, PaddingBits, CString)
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal, filesizeHandler
from hachoir.core.tools import paddingSize


class TOC(FieldSet):
    TYPE_NAME = {
        0x00000001: "Properties",
        0x00000002: "Accelerators",
        0x00000004: "Metrics",
        0x00000008: "Bitmaps",
        0x00000010: "Ink metrics",
        0x00000020: "BDF encodings",
        0x00000040: "SWidths",
        0x00000080: "Glyph names",
        0x00000100: "BDF accelerators",
    }

    FORMAT_NAME = {
        0x00000000: "Default",
        0x00000200: "Ink bounds",
        0x00000100: "Accelerator W ink bounds",
        #        0x00000200: "Compressed metrics",
    }

    def createFields(self):
        yield Enum(UInt32(self, "type"), self.TYPE_NAME)
        yield UInt32(self, "format")
        yield filesizeHandler(UInt32(self, "size"))
        yield UInt32(self, "offset")

    def createDescription(self):
        return "%s at %s (%s)" % (
            self["type"].display, self["offset"].value, self["size"].display)


class PropertiesFormat(FieldSet):
    static_size = 32
    endian = LITTLE_ENDIAN

    def createFields(self):
        yield Bits(self, "reserved[]", 2)
        yield Bit(self, "byte_big_endian")
        yield Bit(self, "bit_big_endian")
        yield Bits(self, "scan_unit", 2)
        yield textHandler(PaddingBits(self, "reserved[]", 26), hexadecimal)


class Property(FieldSet):

    def createFields(self):
        yield UInt32(self, "name_offset")
        yield UInt8(self, "is_string")
        yield UInt32(self, "value_offset")

    def createDescription(self):
        # FIXME: Use link or any better way to read name value
        name = self["../name[%s]" % (self.index - 2)].value
        return "Property %s" % name


class GlyphNames(FieldSet):

    def __init__(self, parent, name, toc, description, size=None):
        FieldSet.__init__(self, parent, name, description, size=size)
        self.toc = toc
        if self["format/byte_big_endian"].value:
            self.endian = BIG_ENDIAN
        else:
            self.endian = LITTLE_ENDIAN

    def createFields(self):
        yield PropertiesFormat(self, "format")
        yield UInt32(self, "count")
        offsets = []
        for index in range(self["count"].value):
            offset = UInt32(self, "offset[]")
            yield offset
            offsets.append(offset.value)
        yield UInt32(self, "total_str_length")
        offsets.sort()
        offset0 = self.current_size // 8
        for offset in offsets:
            padding = self.seekByte(offset0 + offset)
            if padding:
                yield padding
            yield CString(self, "name[]")
        padding = (self.size - self.current_size) // 8
        if padding:
            yield NullBytes(self, "end_padding", padding)


class Properties(GlyphNames):

    def createFields(self):
        yield PropertiesFormat(self, "format")
        yield UInt32(self, "nb_prop")
        properties = []
        for index in range(self["nb_prop"].value):
            property = Property(self, "property[]")
            yield property
            properties.append(property)
        padding = paddingSize(self.current_size // 8, 4)
        if padding:
            yield NullBytes(self, "padding", padding)
        yield UInt32(self, "total_str_length")
        properties.sort(key=lambda entry: entry["name_offset"].value)
        offset0 = self.current_size // 8
        for property in properties:
            padding = self.seekByte(offset0 + property["name_offset"].value)
            if padding:
                yield padding
            yield CString(self, "name[]", "Name of %s" % property.name)
            if property["is_string"].value:
                yield CString(self, "value[]", "Value of %s" % property.name)
        padding = (self.size - self.current_size) // 8
        if padding:
            yield NullBytes(self, "end_padding", padding)


class PcfFile(Parser):
    MAGIC = b"\1fcp"
    PARSER_TAGS = {
        "id": "pcf",
        "category": "misc",
        "file_ext": ("pcf",),
        "magic": ((MAGIC, 0),),
        "min_size": 32,  # FIXME
        "description": "X11 Portable Compiled Font (pcf)",
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        if self["signature"].value != self.MAGIC:
            return "Invalid signature"
        return True

    def createFields(self):
        yield Bytes(self, "signature", 4, r'File signature ("\1pcf")')
        yield UInt32(self, "nb_toc")
        entries = []
        for index in range(self["nb_toc"].value):
            entry = TOC(self, "toc[]")
            yield entry
            entries.append(entry)
        entries.sort(key=lambda entry: entry["offset"].value)
        for entry in entries:
            size = entry["size"].value
            padding = self.seekByte(entry["offset"].value)
            if padding:
                yield padding
            maxsize = (self.size - self.current_size) // 8
            if maxsize < size:
                self.warning("Truncate content of %s to %s bytes (was %s)" % (
                    entry.path, maxsize, size))
                size = maxsize
            if not size:
                continue
            if entry["type"].value == 1:
                yield Properties(self, "properties", entry, "Properties", size=size * 8)
            elif entry["type"].value == 128:
                yield GlyphNames(self, "glyph_names", entry, "Glyph names", size=size * 8)
            else:
                yield RawBytes(self, "data[]", size, "Content of %s" % entry.path)



from hachoir.field import Bit, Bits, FieldSet
from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN
import struct

# Make sure that we use right struct types
assert struct.calcsize("f") == 4
assert struct.calcsize("d") == 8
assert struct.unpack("<d", b"\x1f\x85\xebQ\xb8\x1e\t@")[0] == 3.14
assert struct.unpack(">d", b"\xc0\0\0\0\0\0\0\0")[0] == -2.0


class FloatMantissa(Bits):

    def createValue(self):
        value = Bits.createValue(self)
        return 1 + float(value) / (2 ** self.size)

    def createRawDisplay(self):
        return str(Bits.createValue(self))


class FloatExponent(Bits):

    def __init__(self, parent, name, size):
        Bits.__init__(self, parent, name, size)
        self.bias = 2 ** (size - 1) - 1

    def createValue(self):
        return Bits.createValue(self) - self.bias

    def createRawDisplay(self):
        return str(self.value + self.bias)


def floatFactory(name, format, mantissa_bits, exponent_bits, doc):
    size = 1 + mantissa_bits + exponent_bits

    class Float(FieldSet):
        static_size = size
        __doc__ = doc

        def __init__(self, parent, name, description=None):
            assert parent.endian in (BIG_ENDIAN, LITTLE_ENDIAN)
            FieldSet.__init__(self, parent, name, description, size)
            if format:
                if self._parent.endian == BIG_ENDIAN:
                    self.struct_format = ">" + format
                else:
                    self.struct_format = "<" + format
            else:
                self.struct_format = None

        def createValue(self):
            """
            Create float value: use struct.unpack() when it's possible
            (32 and 64-bit float) or compute it with :
               mantissa * (2.0 ** exponent)

            This computation may raise an OverflowError.
            """
            if self.struct_format:
                raw = self._parent.stream.readBytes(
                    self.absolute_address, self._size // 8)
                try:
                    return struct.unpack(self.struct_format, raw)[0]
                except struct.error as err:
                    raise ValueError("[%s] conversion error: %s" %
                                     (self.__class__.__name__, err))
            else:
                try:
                    value = self["mantissa"].value * \
                        (2.0 ** float(self["exponent"].value))
                    if self["negative"].value:
                        return -(value)
                    else:
                        return value
                except OverflowError:
                    raise ValueError("[%s] floating point overflow" %
                                     self.__class__.__name__)

        def createFields(self):
            yield Bit(self, "negative")
            yield FloatExponent(self, "exponent", exponent_bits)
            if 64 <= mantissa_bits:
                yield Bit(self, "one")
                yield FloatMantissa(self, "mantissa", mantissa_bits - 1)
            else:
                yield FloatMantissa(self, "mantissa", mantissa_bits)

    cls = Float
    cls.__name__ = name
    return cls


# 32-bit float (standard: IEEE 754/854)
Float32 = floatFactory("Float32", "f", 23, 8,
                       "Floating point number: format IEEE 754 int 32 bit")

# 64-bit float (standard: IEEE 754/854)
Float64 = floatFactory("Float64", "d", 52, 11,
                       "Floating point number: format IEEE 754 in 64 bit")

# 80-bit float (standard: IEEE 754/854)
Float80 = floatFactory("Float80", None, 64, 15,
                       "Floating point number: format IEEE 754 in 80 bit")



"""
String field classes:
- String: Fixed length string (no prefix/no suffix) ;
- CString: String which ends with nul byte ("\0") ;
- UnixLine: Unix line of text, string which ends with "\n" ;
- PascalString8, PascalString16, PascalString32: String prefixed with
  length written in a 8, 16, 32-bit integer (use parent endian).

Constructor has optional arguments:
- strip: value can be a string or True ;
- charset: if set, convert string to unicode using this charset (in "replace"
  mode which replace all buggy characters with ".").

Note: For PascalStringXX, prefixed value is the number of bytes and not
      of characters!
"""

from hachoir.field import FieldError, Bytes
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.tools import alignValue, makePrintable
from hachoir.core.i18n import guessBytesCharset
from hachoir.core import config
from codecs import BOM_UTF16_LE, BOM_UTF16_BE, BOM_UTF32_LE, BOM_UTF32_BE

# Default charset used to convert byte string to Unicode
# This charset is used if no charset is specified or on conversion error
FALLBACK_CHARSET = "ISO-8859-1"


class GenericString(Bytes):
    """
    Generic string class.

    charset have to be in CHARSET_8BIT or in UTF_CHARSET.
    """

    VALID_FORMATS = ("C", "UnixLine",
                     "fixed", "Pascal8", "Pascal16", "Pascal32")

    # 8-bit charsets
    CHARSET_8BIT = set((
        "ASCII",          # ANSI X3.4-1968
        "MacRoman",
        "CP037",          # EBCDIC 037
        "CP874",          # Thai
        "WINDOWS-1250",   # Central Europe
        "WINDOWS-1251",   # Cyrillic
        "WINDOWS-1252",   # Latin I
        "WINDOWS-1253",   # Greek
        "WINDOWS-1254",   # Turkish
        "WINDOWS-1255",   # Hebrew
        "WINDOWS-1256",   # Arabic
        "WINDOWS-1257",   # Baltic
        "WINDOWS-1258",   # Vietnam
        "ISO-8859-1",     # Latin-1
        "ISO-8859-2",     # Latin-2
        "ISO-8859-3",     # Latin-3
        "ISO-8859-4",     # Latin-4
        "ISO-8859-5",
        "ISO-8859-6",
        "ISO-8859-7",
        "ISO-8859-8",
        "ISO-8859-9",     # Latin-5
        "ISO-8859-10",    # Latin-6
        "ISO-8859-11",    # Thai
        "ISO-8859-13",    # Latin-7
        "ISO-8859-14",    # Latin-8
        "ISO-8859-15",    # Latin-9 or ("Latin-0")
        "ISO-8859-16",    # Latin-10
    ))

    # UTF-xx charset familly
    UTF_CHARSET = {
        "UTF-8": (8, None),
        "UTF-16-LE": (16, LITTLE_ENDIAN),
        "UTF-32LE": (32, LITTLE_ENDIAN),
        "UTF-16-BE": (16, BIG_ENDIAN),
        "UTF-32BE": (32, BIG_ENDIAN),
        "UTF-16": (16, "BOM"),
        "UTF-32": (32, "BOM"),
    }

    # UTF-xx BOM => charset with endian
    UTF_BOM = {
        16: {BOM_UTF16_LE: "UTF-16-LE", BOM_UTF16_BE: "UTF-16-BE"},
        32: {BOM_UTF32_LE: "UTF-32LE", BOM_UTF32_BE: "UTF-32BE"},
    }

    # Suffix format: value is suffix (string)
    SUFFIX_FORMAT = {
        "C": {
            8: {LITTLE_ENDIAN: b"\0", BIG_ENDIAN: b"\0"},
            16: {LITTLE_ENDIAN: b"\0\0", BIG_ENDIAN: b"\0\0"},
            32: {LITTLE_ENDIAN: b"\0\0\0\0", BIG_ENDIAN: b"\0\0\0\0"},
        },
        "UnixLine": {
            8: {LITTLE_ENDIAN: b"\n", BIG_ENDIAN: b"\n"},
            16: {LITTLE_ENDIAN: b"\n\0", BIG_ENDIAN: b"\0\n"},
            32: {LITTLE_ENDIAN: b"\n\0\0\0", BIG_ENDIAN: b"\0\0\0\n"},
        },

    }

    # Pascal format: value is the size of the prefix in bits
    PASCAL_FORMATS = {
        "Pascal8": 1,
        "Pascal16": 2,
        "Pascal32": 4
    }

    # Raw value: with prefix and suffix, not stripped,
    # and not converted to Unicode
    _raw_value = None

    def __init__(self, parent, name, format, description=None,
                 strip=None, charset=None, nbytes=None, truncate=None):
        Bytes.__init__(self, parent, name, 1, description)

        # Is format valid?
        assert format in self.VALID_FORMATS

        # Store options
        self._format = format
        self._strip = strip
        self._truncate = truncate

        # Check charset and compute character size in bytes
        # (or None when it's not possible to guess character size)
        if not charset or charset in self.CHARSET_8BIT:
            self._character_size = 1   # one byte per character
        elif charset in self.UTF_CHARSET:
            self._character_size = None
        else:
            raise FieldError("Invalid charset for %s: \"%s\"" %
                             (self.path, charset))
        self._charset = charset

        # It is a fixed string?
        if nbytes is not None:
            assert self._format == "fixed"
            # Arbitrary limits, just to catch some bugs...
            if not (1 <= nbytes <= 0xffff):
                raise FieldError("Invalid string size for %s: %s" %
                                 (self.path, nbytes))
            self._content_size = nbytes   # content length in bytes
            self._size = nbytes * 8
            self._content_offset = 0
        else:
            # Format with a suffix: Find the end of the string
            if self._format in self.SUFFIX_FORMAT:
                self._content_offset = 0

                # Choose the suffix
                suffix = self.suffix_str

                # Find the suffix
                length = self._parent.stream.searchBytesLength(
                    suffix, False, self.absolute_address)
                if length is None:
                    raise FieldError("Unable to find end of string %s (format %s)!"
                                     % (self.path, self._format))
                if 1 < len(suffix):
                    # Fix length for little endian bug with UTF-xx charset:
                    #   u"abc" -> "a\0b\0c\0\0\0" (UTF-16-LE)
                    #   search returns length=5, whereas real lenght is 6
                    length = alignValue(length, len(suffix))

                # Compute sizes
                self._content_size = length  # in bytes
                self._size = (length + len(suffix)) * 8

            # Format with a prefix: Read prefixed length in bytes
            else:
                assert self._format in self.PASCAL_FORMATS

                # Get the prefix size
                prefix_size = self.PASCAL_FORMATS[self._format]
                self._content_offset = prefix_size

                # Read the prefix and compute sizes
                value = self._parent.stream.readBits(
                    self.absolute_address, prefix_size * 8, self._parent.endian)
                self._content_size = value   # in bytes
                self._size = (prefix_size + value) * 8

        # For UTF-16 and UTF-32, choose the right charset using BOM
        if self._charset in self.UTF_CHARSET:
            # Charset requires a BOM?
            bomsize, endian = self.UTF_CHARSET[self._charset]
            if endian == "BOM":
                # Read the BOM value
                nbytes = bomsize // 8
                bom = self._parent.stream.readBytes(
                    self.absolute_address, nbytes)

                # Choose right charset using the BOM
                bom_endian = self.UTF_BOM[bomsize]
                if bom not in bom_endian:
                    raise FieldError("String %s has invalid BOM (%s)!"
                                     % (self.path, repr(bom)))
                self._charset = bom_endian[bom]
                self._content_size -= nbytes
                self._content_offset += nbytes

        # Compute length in character if possible
        if self._character_size:
            self._length = self._content_size // self._character_size
        else:
            self._length = None

    @staticmethod
    def staticSuffixStr(format, charset, endian):
        if format not in GenericString.SUFFIX_FORMAT:
            return ''
        suffix = GenericString.SUFFIX_FORMAT[format]
        if charset in GenericString.UTF_CHARSET:
            suffix_size = GenericString.UTF_CHARSET[charset][0]
            suffix = suffix[suffix_size]
        else:
            suffix = suffix[8]
        return suffix[endian]

    def _getSuffixStr(self):
        return self.staticSuffixStr(
            self._format, self._charset, self._parent.endian)
    suffix_str = property(_getSuffixStr)

    def _convertText(self, text):
        if not self._charset:
            # charset is still unknown: guess the charset
            self._charset = guessBytesCharset(text, default=FALLBACK_CHARSET)

        # Try to convert to Unicode
        try:
            return str(text, self._charset, "strict")
        except UnicodeDecodeError as exc:
            err = exc

        # --- Conversion error ---

        # Fix truncated UTF-16 string like 'B\0e' (3 bytes)
        # => Add missing nul byte: 'B\0e\0' (4 bytes)
        if err.reason == "truncated data" \
                and err.end == len(text) \
                and self._charset == "UTF-16-LE":
            try:
                text = str(text + b"\0", self._charset, "strict")
                self.warning(
                    "Fix truncated %s string: add missing nul byte" % self._charset)
                return text
            except UnicodeDecodeError:
                pass

        # On error, use FALLBACK_CHARSET
        self.warning("Unable to convert string to Unicode: %s" % err)
        return str(text, FALLBACK_CHARSET, "strict")

    def _guessCharset(self):
        addr = self.absolute_address + self._content_offset * 8
        bytes = self._parent.stream.readBytes(addr, self._content_size)
        return guessBytesCharset(bytes, default=FALLBACK_CHARSET)

    def createValue(self, human=True):
        # Compress data address (in bits) and size (in bytes)
        if human:
            addr = self.absolute_address + self._content_offset * 8
            size = self._content_size
        else:
            addr = self.absolute_address
            size = self._size // 8
        if size == 0:
            # Empty string
            return ""

        # Read bytes in data stream
        text = self._parent.stream.readBytes(addr, size)

        # Don't transform data?
        if not human:
            return text

        # Convert text to Unicode
        text = self._convertText(text)

        # Truncate
        if self._truncate:
            pos = text.find(self._truncate)
            if 0 <= pos:
                text = text[:pos]

        # Strip string if needed
        if self._strip:
            if isinstance(self._strip, str):
                text = text.strip(self._strip)
            else:
                text = text.strip()
        assert isinstance(text, str)
        return text

    def createDisplay(self, human=True):
        if not human:
            if self._raw_value is None:
                self._raw_value = GenericString.createValue(self, False)
            value = makePrintable(self._raw_value, "ASCII")
        elif self._charset:
            value = makePrintable(self.value, "ISO-8859-1")
        else:
            value = self.value
        if config.max_string_length < len(value):
            # Truncate string if needed
            value = "%s(...)" % value[:config.max_string_length]
        if not self._charset or not human:
            return makePrintable(value, "ASCII", quote='"')
        else:
            if value:
                return '"%s"' % value.replace('"', '\\"')
            else:
                return "(empty)"

    def createRawDisplay(self):
        return GenericString.createDisplay(self, human=False)

    def _getLength(self):
        if self._length is None:
            self._length = len(self.value)
        return self._length
    length = property(_getLength, doc="String length in characters")

    def _getFormat(self):
        return self._format
    format = property(_getFormat, doc="String format (eg. 'C')")

    def _getCharset(self):
        if not self._charset:
            self._charset = self._guessCharset()
        return self._charset
    charset = property(_getCharset, doc="String charset (eg. 'ISO-8859-1')")

    def _getContentSize(self):
        return self._content_size
    content_size = property(_getContentSize, doc="Content size in bytes")

    def _getContentOffset(self):
        return self._content_offset
    content_offset = property(_getContentOffset, doc="Content offset in bytes")

    def getFieldType(self):
        info = self.charset
        if self._strip:
            if isinstance(self._strip, str):
                info += ",strip=%s" % makePrintable(
                    self._strip, "ASCII", quote="'")
            else:
                info += ",strip=True"
        return "%s<%s>" % (Bytes.getFieldType(self), info)


def stringFactory(name, format, doc):
    class NewString(GenericString):
        __doc__ = doc

        def __init__(self, parent, name, description=None,
                     strip=None, charset=None, truncate=None):
            GenericString.__init__(self, parent, name, format, description,
                                   strip=strip, charset=charset, truncate=truncate)
    cls = NewString
    cls.__name__ = name
    return cls


# String which ends with nul byte ("\0")
CString = stringFactory("CString", "C",
                        r"""C string: string ending with nul byte.
See GenericString to get more information.""")

# Unix line of text: string which ends with "\n" (ASCII 0x0A)
UnixLine = stringFactory("UnixLine", "UnixLine",
                         r"""Unix line: string ending with "\n" (ASCII code 10).
See GenericString to get more information.""")

# String prefixed with length written in a 8-bit integer
PascalString8 = stringFactory("PascalString8", "Pascal8",
                              r"""Pascal string: string prefixed with 8-bit integer containing its length (endian depends on parent endian).
See GenericString to get more information.""")

# String prefixed with length written in a 16-bit integer (use parent endian)
PascalString16 = stringFactory("PascalString16", "Pascal16",
                               r"""Pascal string: string prefixed with 16-bit integer containing its length (endian depends on parent endian).
See GenericString to get more information.""")

# String prefixed with length written in a 32-bit integer (use parent endian)
PascalString32 = stringFactory("PascalString32", "Pascal32",
                               r"""Pascal string: string prefixed with 32-bit integer containing its length (endian depends on parent endian).
See GenericString to get more information.""")


class String(GenericString):
    """
    String with fixed size (size in bytes).
    See GenericString to get more information.
    """
    static_size = staticmethod(lambda *args, **kw: args[1] * 8)

    def __init__(self, parent, name, nbytes, description=None,
                 strip=None, charset=None, truncate=None):
        GenericString.__init__(self, parent, name, "fixed", description,
                               strip=strip, charset=charset, nbytes=nbytes, truncate=truncate)


String.__name__ = "FixedString"



"""
Dictionnary classes which store values order.
"""


class UniqKeyError(Exception):
    """
    Error raised when a value is set whereas the key already exist in a
    dictionnary.
    """
    pass


class Dict(object):
    """
    This class works like classic Python dict() but has an important method:
    __iter__() which allow to iterate into the dictionnary _values_ (and not
    keys like Python's dict does).
    """

    def __init__(self, values=None):
        self._index = {}        # key => index
        self._key_list = []     # index => key
        self._value_list = []   # index => value
        if values:
            for key, value in values:
                self.append(key, value)

    def _getValues(self):
        return self._value_list
    values = property(_getValues)

    def index(self, key):
        """
        Search a value by its key and returns its index
        Returns None if the key doesn't exist.

        >>> d=Dict( (("two", "deux"), ("one", "un")) )
        >>> d.index("two")
        0
        >>> d.index("one")
        1
        >>> d.index("three") is None
        True
        """
        return self._index.get(key)

    def __getitem__(self, key):
        """
        Get item with specified key.
        To get a value by it's index, use mydict.values[index]

        >>> d=Dict( (("two", "deux"), ("one", "un")) )
        >>> d["one"]
        'un'
        """
        return self._value_list[self._index[key]]

    def __setitem__(self, key, value):
        self._value_list[self._index[key]] = value

    def append(self, key, value):
        """
        Append new value
        """
        if key in self._index:
            raise UniqKeyError("Key '%s' already exists" % key)
        self._index[key] = len(self._value_list)
        self._key_list.append(key)
        self._value_list.append(value)

    def __len__(self):
        return len(self._value_list)

    def __contains__(self, key):
        return key in self._index

    def __iter__(self):
        return iter(self._value_list)

    def items(self):
        """
        Create a generator to iterate on: (key, value).

        >>> d=Dict( (("two", "deux"), ("one", "un")) )
        >>> for key, value in d.items():
        ...    print("%r: %r" % (key, value))
        ...
        'two': 'deux'
        'one': 'un'
        """
        for index in range(len(self)):
            yield (self._key_list[index], self._value_list[index])

    def itervalues(self):
        """
        Create an iterator on values
        """
        return iter(self._value_list)

    def iterkeys(self):
        """
        Create an iterator on keys
        """
        return iter(self._key_list)

    def replace(self, oldkey, newkey, new_value):
        """
        Replace an existing value with another one

        >>> d=Dict( (("two", "deux"), ("one", "un")) )
        >>> d.replace("one", "three", 3)
        >>> d
        {'two': 'deux', 'three': 3}

        You can also use the classic form:

        >>> d['three'] = 4
        >>> d
        {'two': 'deux', 'three': 4}
        """
        index = self._index[oldkey]
        self._value_list[index] = new_value
        if oldkey != newkey:
            del self._index[oldkey]
            self._index[newkey] = index
            self._key_list[index] = newkey

    def __delitem__(self, index):
        """
        Delete item at position index. May raise IndexError.

        >>> d=Dict( ((6, 'six'), (9, 'neuf'), (4, 'quatre')) )
        >>> del d[1]
        >>> d
        {6: 'six', 4: 'quatre'}
        """
        if index < 0:
            index += len(self._value_list)
        if not (0 <= index < len(self._value_list)):
            raise IndexError("list assignment index out of range (%s/%s)"
                             % (index, len(self._value_list)))
        del self._value_list[index]
        del self._key_list[index]

        # First loop which may alter self._index
        for key, item_index in self._index.items():
            if item_index == index:
                del self._index[key]
                break

        # Second loop update indexes
        for key, item_index in self._index.items():
            if index < item_index:
                self._index[key] -= 1

    def insert(self, index, key, value):
        """
        Insert an item at specified position index.

        >>> d=Dict( ((6, 'six'), (9, 'neuf'), (4, 'quatre')) )
        >>> d.insert(1, '40', 'quarante')
        >>> d
        {6: 'six', '40': 'quarante', 9: 'neuf', 4: 'quatre'}
        """
        if key in self:
            raise UniqKeyError("Insert error: key '%s' ready exists" % key)
        _index = index
        if index < 0:
            index += len(self._value_list)
        if not (0 <= index <= len(self._value_list)):
            raise IndexError("Insert error: index '%s' is invalid" % _index)
        for item_key, item_index in self._index.items():
            if item_index >= index:
                self._index[item_key] += 1
        self._index[key] = index
        self._key_list.insert(index, key)
        self._value_list.insert(index, value)

    def __repr__(self):
        items = ("%r: %r" % (key, value) for key, value in self.items())
        return "{%s}" % ", ".join(items)



"""
FLV video parser.

Documentation:

 - FLV File format: http://osflash.org/flv
 - libavformat from ffmpeg project
 - flashticle: Python project to read Flash (SWF and FLV with AMF metadata)
   http://undefined.org/python/#flashticle

Author: Victor Stinner
Creation date: 4 november 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt8, UInt24, UInt32, NullBits, NullBytes,
                           Bit, Bits, String, RawBytes, Enum)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.parser.audio.mpeg_audio import Frame
from hachoir.parser.video.amf import AMFObject
from hachoir.core.tools import createDict

SAMPLING_RATE = {
    0: (5512, "5.5 kHz"),
    1: (11025, "11 kHz"),
    2: (22050, "22.1 kHz"),
    3: (44100, "44.1 kHz"),
}
SAMPLING_RATE_VALUE = createDict(SAMPLING_RATE, 0)
SAMPLING_RATE_TEXT = createDict(SAMPLING_RATE, 1)

AUDIO_CODEC_MP3 = 2
AUDIO_CODEC_NAME = {
    0: "Uncompressed",
    1: "ADPCM",
    2: "MP3",
    5: "Nellymoser 8kHz mono",
    6: "Nellymoser",
}

VIDEO_CODEC_NAME = {
    2: "Sorensen H.263",
    3: "Screen video",
    4: "On2 VP6",
}

FRAME_TYPE = {
    1: "keyframe",
    2: "inter frame",
    3: "disposable inter frame",
}


class Header(FieldSet):

    def createFields(self):
        yield String(self, "signature", 3, "FLV format signature", charset="ASCII")
        yield UInt8(self, "version")

        yield NullBits(self, "reserved[]", 5)
        yield Bit(self, "type_flags_audio")
        yield NullBits(self, "reserved[]", 1)
        yield Bit(self, "type_flags_video")

        yield UInt32(self, "data_offset")


def parseAudio(parent, size):
    yield Enum(Bits(parent, "codec", 4, "Audio codec"), AUDIO_CODEC_NAME)
    yield Enum(Bits(parent, "sampling_rate", 2, "Sampling rate"), SAMPLING_RATE_TEXT)
    yield Bit(parent, "is_16bit", "16-bit or 8-bit per sample")
    yield Bit(parent, "is_stereo", "Stereo or mono channel")

    size -= 1
    if 0 < size:
        if parent["codec"].value == AUDIO_CODEC_MP3:
            yield Frame(parent, "music_data", size=size * 8)
        else:
            yield RawBytes(parent, "music_data", size)


def parseVideo(parent, size):
    yield Enum(Bits(parent, "frame_type", 4, "Frame type"), FRAME_TYPE)
    yield Enum(Bits(parent, "codec", 4, "Video codec"), VIDEO_CODEC_NAME)
    if 1 < size:
        yield RawBytes(parent, "data", size - 1)


def parseAMF(parent, size):
    while parent.current_size < parent.size:
        yield AMFObject(parent, "entry[]")


class Chunk(FieldSet):
    tag_info = {
        8: ("audio[]", parseAudio, ""),
        9: ("video[]", parseVideo, ""),
        18: ("metadata", parseAMF, ""),
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = (11 + self["size"].value) * 8
        tag = self["tag"].value
        if tag in self.tag_info:
            self._name, self.parser, self._description = self.tag_info[tag]
        else:
            self.parser = None

    def createFields(self):
        yield UInt8(self, "tag")
        yield UInt24(self, "size", "Content size")
        yield UInt24(self, "timestamp", "Timestamp in millisecond")
        yield NullBytes(self, "reserved", 4)
        size = self["size"].value
        if size:
            if self.parser:
                yield from self.parser(self, size)
            else:
                yield RawBytes(self, "content", size)

    def getSampleRate(self):
        try:
            return SAMPLING_RATE_VALUE[self["sampling_rate"].value]
        except LookupError:
            return None


class FlvFile(Parser):
    PARSER_TAGS = {
        "id": "flv",
        "category": "video",
        "file_ext": ("flv",),
        "mime": ("video/x-flv",),
        "min_size": 9 * 4,
        "magic": (
            # Signature, version=1, flags=5 (video+audio), header size=9
            (b"FLV\1\x05\0\0\0\x09", 0),
            # Signature, version=1, flags=5 (video), header size=9
            (b"FLV\1\x01\0\0\0\x09", 0),
        ),
        "description": "Macromedia Flash video"
    }
    endian = BIG_ENDIAN

    def validate(self):
        if self.stream.readBytes(0, 3) != b"FLV":
            return "Wrong file signature"
        if self["header/data_offset"].value != 9:
            return "Unknown data offset in main header"
        return True

    def createFields(self):
        yield Header(self, "header")
        yield UInt32(self, "prev_size[]", "Size of previous chunk")
        while not self.eof:
            yield Chunk(self, "chunk[]")
            yield UInt32(self, "prev_size[]", "Size of previous chunk")

    def createDescription(self):
        return "Macromedia Flash video version %s" % self["header/version"].value



from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN  # noqa
from hachoir.stream.stream import StreamError  # noqa
from hachoir.stream.input import (InputStreamError,  # noqa
                                  InputStream, InputIOStream, StringInputStream,
                                  InputSubStream, InputFieldStream,
                                  FragmentedStream, ConcatStream)
from hachoir.stream.input_helper import FileInputStream, guessStreamCharset  # noqa
from hachoir.stream.output import (OutputStreamError,  # noqa
                                   FileOutputStream, StringOutputStream, OutputStream)



"""
Configuration of Hachoir
"""

import os

# UI: display options
max_string_length = 40    # Max. length in characters of GenericString.display
max_byte_length = 14      # Max. length in bytes of RawBytes.display
max_bit_length = 256      # Max. length in bits of RawBits.display

# Global options
debug = False             # Display many informations usefull to debug
verbose = False           # Display more informations
quiet = False             # Don't display warnings

# Use internationalization and localization (gettext)?
if os.name == "nt":
    # TODO: Remove this hack and make i18n works on Windows :-)
    use_i18n = False
else:
    use_i18n = True

# Parser global options
autofix = True            # Enable Autofix? see hachoir.field.GenericFieldSet
check_padding_pattern = True   # Check padding fields pattern?
